#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Module for testing the satpy.readers.core.hdf4 module."""

import os
import unittest

import numpy as np
import xarray as xr

try:
    from satpy.readers.core.hdf4 import HDF4FileHandler
except ImportError:
    # fake the import so we can at least run the tests in this file
    HDF4FileHandler = object  # type: ignore


class FakeHDF4FileHandler(HDF4FileHandler):
    """Swap-in NetCDF4 File Handler for reader tests to use."""

    def __init__(self, filename, filename_info, filetype_info, **kwargs):
        """Get fake file content from 'get_test_content'."""
        if HDF4FileHandler is object:
            raise ImportError("Base 'HDF4FileHandler' could not be "
                              "imported.")
        super(HDF4FileHandler, self).__init__(filename, filename_info, filetype_info)
        self.file_content = self.get_test_content(filename, filename_info, filetype_info)
        self.file_content.update(kwargs)

    def get_test_content(self, filename, filename_info, filetype_info):
        """Mimic reader input file content.

        Args:
            filename (str): input filename
            filename_info (dict): Dict of metadata pulled from filename
            filetype_info (dict): Dict of metadata from the reader's yaml config for this file type

        Returns: dict of file content with keys like:

            - 'dataset'
            - '/attr/global_attr'
            - 'dataset/attr/global_attr'
            - 'dataset/shape'

        """
        raise NotImplementedError("Fake File Handler subclass must implement 'get_test_content'")


class TestHDF4FileHandler(unittest.TestCase):
    """Test HDF4 File Handler Utility class."""

    def setUp(self):
        """Create a test HDF4 file."""
        from pyhdf.SD import SD, SDC
        h = SD("test.hdf", SDC.WRITE | SDC.CREATE | SDC.TRUNC)
        data = np.arange(10. * 100, dtype=np.float32).reshape((10, 100))
        v1 = h.create("ds1_f", SDC.FLOAT32, (10, 100))
        v1[:] = data
        v2 = h.create("ds1_i", SDC.INT16, (10, 100))
        v2[:] = data.astype(np.int16)

        # Add attributes
        h.test_attr_str = "test_string"
        h.test_attr_int = 0
        h.test_attr_float = 1.2
        # h.test_attr_str_arr = np.array(b"test_string2")
        for d in [v1, v2]:
            d.test_attr_str = "test_string"
            d.test_attr_int = 0
            d.test_attr_float = 1.2

        h.end()

    def tearDown(self):
        """Remove the previously created test file."""
        os.remove("test.hdf")

    def test_all_basic(self):
        """Test everything about the HDF4 class."""
        from satpy.readers.core.hdf4 import HDF4FileHandler
        file_handler = HDF4FileHandler("test.hdf", {}, {})

        for ds in ("ds1_f", "ds1_i"):
            assert file_handler[ds + "/dtype"] == (np.float32 if ds.endswith("f") else np.int16)
            assert file_handler[ds + "/shape"] == (10, 100)
            # make sure that the dtype is an instance, not the class
            assert file_handler[ds].dtype.itemsize == (4 if ds.endswith("f") else 2)
            attrs = file_handler[ds].attrs
            assert attrs.get("test_attr_str") == "test_string"
            assert attrs.get("test_attr_int") == 0
            assert attrs.get("test_attr_float") == 1.2

        assert isinstance(file_handler["/attr/test_attr_str"], str)
        assert file_handler["/attr/test_attr_str"] == "test_string"
        # self.assertEqual(file_handler['/attr/test_attr_str_arr'], 'test_string2')
        assert isinstance(file_handler["/attr/test_attr_int"], int)
        assert file_handler["/attr/test_attr_int"] == 0
        assert isinstance(file_handler["/attr/test_attr_float"], float)
        assert file_handler["/attr/test_attr_float"] == 1.2

        assert isinstance(file_handler.get("ds1_f"), xr.DataArray)
        assert file_handler.get("fake_ds") is None
        assert file_handler.get("fake_ds", "test") == "test"

        assert "ds1_f" in file_handler
        assert "fake_ds" not in file_handler



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2011-2015 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Interface to OMPS EDR format."""

import datetime as dt
import logging

import numpy as np

from satpy.readers.core.hdf5 import HDF5FileHandler

NO_DATE = dt.datetime(1958, 1, 1)
EPSILON_TIME = dt.timedelta(days=2)
LOG = logging.getLogger(__name__)


class EDRFileHandler(HDF5FileHandler):
    """EDR file handler."""

    _fill_name = "_FillValue"

    @property
    def start_orbit_number(self):
        """Get the start orbit number."""
        return self.filename_info["orbit"]

    @property
    def end_orbit_number(self):
        """Get the end orbit number."""
        return self.filename_info["orbit"]

    @property
    def platform_name(self):
        """Get the platform name."""
        return self.filename_info["platform_shortname"]

    @property
    def sensor_name(self):
        """Get the sensor name."""
        return self.filename_info["instrument_shortname"]

    def get_shape(self, ds_id, ds_info):
        """Get the shape."""
        return self[ds_info["file_key"] + "/shape"]

    def adjust_scaling_factors(self, factors, file_units, output_units):
        """Adjust scaling factors."""
        if factors is None or factors[0] is None:
            factors = [1, 0]
        if file_units == output_units:
            LOG.debug("File units and output units are the same (%s)", file_units)
            return factors
        return np.array(factors)

    def get_metadata(self, dataset_id, ds_info):
        """Get the metadata."""
        var_path = ds_info.get("file_key", "{}".format(dataset_id["name"]))
        info = getattr(self[var_path], "attrs", {}).copy()
        info.pop("DIMENSION_LIST", None)
        info.update(ds_info)

        file_units = ds_info.get("file_units")
        if file_units is None:
            file_units = self.get(var_path + "/attr/units", self.get(var_path + "/attr/Units"))
        if file_units is None:
            raise KeyError("File variable '{}' has no units attribute".format(var_path))
        if file_units == "deg":
            file_units = "degrees"
        elif file_units == "Unitless":
            file_units = "1"

        info.update({
            "shape": self.get_shape(dataset_id, ds_info),
            "file_units": file_units,
            "units": ds_info.get("units", file_units),
            "platform_name": self.platform_name,
            "sensor": self.sensor_name,
            "start_orbit": self.start_orbit_number,
            "end_orbit": self.end_orbit_number,
        })
        info.update(dataset_id.to_dict())
        if "standard_name" not in ds_info:
            info["standard_name"] = self.get(var_path + "/attr/Title", dataset_id["name"])
        return info

    def get_dataset(self, dataset_id, ds_info):
        """Get the dataset."""
        var_path = ds_info.get("file_key", "{}".format(dataset_id["name"]))
        metadata = self.get_metadata(dataset_id, ds_info)
        valid_min, valid_max = self.get(var_path + "/attr/valid_range",
                                        self.get(var_path + "/attr/ValidRange", (None, None)))
        if valid_min is None or valid_max is None:
            valid_min = self.get(var_path + "/attr/valid_min", None)
            valid_max = self.get(var_path + "/attr/valid_max", None)
            if valid_min is None or valid_max is None:
                raise KeyError("File variable '{}' has no valid range attribute".format(var_path))
        fill_name = var_path + "/attr/{}".format(self._fill_name)
        if fill_name in self:
            fill_value = self[fill_name]
        else:
            fill_value = None

        data = self[var_path]
        scale_factor_path = var_path + "/attr/ScaleFactor"
        if scale_factor_path in self:
            scale_factor = self[scale_factor_path]
            scale_offset = self[var_path + "/attr/Offset"]
        else:
            scale_factor = None
            scale_offset = None

        if valid_min is not None and valid_max is not None:
            # the original .cfg/INI based reader only checked valid_max
            data = data.where((data <= valid_max) & (data >= valid_min))
        if fill_value is not None:
            data = data.where(data != fill_value)

        factors = (scale_factor, scale_offset)
        factors = self.adjust_scaling_factors(factors, metadata["file_units"], ds_info.get("units"))
        if factors[0] != 1 or factors[1] != 0:
            data = data * factors[0] + factors[1]

        data.attrs.update(metadata)
        if "DIMENSION_LIST" in data.attrs:
            data.attrs.pop("DIMENSION_LIST")
            dimensions = self.get_reference(var_path, "DIMENSION_LIST")
            for dim, coord in zip(data.dims, dimensions):
                data.coords[dim] = coord[0]
        return data


class EDREOSFileHandler(EDRFileHandler):
    """EDR EOS file handler."""

    _fill_name = "MissingValue"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""The fy4_base reader tests package."""

from unittest import mock

import pytest

from satpy.readers.core.fy4 import FY4Base
from satpy.tests.reader_tests.test_agri_l1 import FakeHDF5FileHandler2


class Test_FY4Base:
    """Tests for the FengYun4 base class for the components missed by AGRI/GHI tests."""

    def setup_method(self):
        """Initialise the tests."""
        self.p = mock.patch.object(FY4Base, "__bases__", (FakeHDF5FileHandler2,))
        self.fake_handler = self.p.start()
        self.p.is_local = True

        self.file_type = {"file_type": "agri_l1_0500m"}

    def teardown_method(self):
        """Stop wrapping the HDF5 file handler."""
        self.p.stop()

    def test_badsensor(self):
        """Test case where we pass a bad sensor name, must be GHI or AGRI."""
        fy4 = FY4Base(None, {"platform_id": "FY4A", "instrument": "FCI"}, self.file_type)
        with pytest.raises(ValueError, match="Unsupported sensor type: FCI"):
            fy4.calibrate_to_reflectance(None, None, None)
        with pytest.raises(ValueError, match="Error, sensor must be GHI or AGRI."):
            fy4.calibrate_to_bt(None, None, None)

    def test_badcalibration(self):
        """Test case where we pass a bad calibration type, radiance is not supported."""
        fy4 = FY4Base(None, {"platform_id": "FY4A", "instrument": "AGRI"}, self.file_type)
        with pytest.raises(NotImplementedError):
            fy4.calibrate(None, {"calibration": "radiance"}, None, None)

    def test_badplatform(self):
        """Test case where we pass a bad calibration type, radiance is not supported."""
        with pytest.raises(KeyError):
            FY4Base(None, {"platform_id": "FY3D", "instrument": "AGRI"}, self.file_type)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2010-2022 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Scene object to hold satellite data."""
from __future__ import annotations

import logging
import os
import warnings
from collections.abc import Iterable
from typing import Any, Callable

import numpy as np
import xarray as xr
from pyresample.geometry import AreaDefinition, BaseDefinition, CoordinateDefinition, SwathDefinition
from xarray import DataArray

from satpy.area import get_area_def
from satpy.composites.config_loader import load_compositor_configs_for_sensors
from satpy.composites.core import IncompatibleAreas
from satpy.dataset import DataID, DataQuery, DatasetDict, combine_metadata, dataset_walker, replace_anc
from satpy.dependency_tree import DependencyTree
from satpy.node import CompositorNode, MissingDependencies, ReaderNode
from satpy.readers.core.loading import load_readers
from satpy.utils import convert_remote_files_to_fsspec, get_storage_options_from_reader_kwargs

LOG = logging.getLogger(__name__)


def _get_area_resolution(area):
    """Attempt to retrieve resolution from AreaDefinition."""
    try:
        resolution = max(area.pixel_size_x, area.pixel_size_y)
    except AttributeError:
        resolution = max(area.lats.attrs["resolution"], area.lons.attrs["resolution"])
    return resolution


def _aggregate_data_array(data_array, func, **coarsen_kwargs):
    """Aggregate xr.DataArray."""
    res = data_array.coarsen(**coarsen_kwargs)
    if callable(func):
        out = res.reduce(func)
    else:
        out = getattr(res, func)()
    return out


class DelayedGeneration(KeyError):
    """Mark that a dataset can't be generated without further modification."""

    pass


class Scene:
    """The Almighty Scene Class.

    Example usage::

        from satpy import Scene
        from glob import glob

        # create readers and open files
        scn = Scene(filenames=glob('/path/to/files/*'), reader='viirs_sdr')

        # load datasets from input files
        scn.load(['I01', 'I02'])

        # resample from satellite native geolocation to builtin 'eurol' Area
        new_scn = scn.resample('eurol')

        # save all resampled datasets to geotiff files in the current directory
        new_scn.save_datasets()

    """

    def __init__(self, filenames=None, reader=None, filter_parameters=None,
                 reader_kwargs=None):
        """Initialize Scene with Reader and Compositor objects.

        To load data `filenames` and preferably `reader` must be specified::

            scn = Scene(filenames=glob('/path/to/viirs/sdr/files/*'), reader='viirs_sdr')


        If ``filenames`` is provided without ``reader`` then the available readers
        will be searched for a Reader that can support the provided files. This
        can take a considerable amount of time so it is recommended that
        ``reader`` always be provided. Note without ``filenames`` the Scene is
        created with no Readers available. When a Scene is created with no Readers,
        each xarray.DataArray must be added manually::

            scn = Scene()
            scn['my_dataset'] = DataArray(my_data_array, attrs={})

        The `attrs` dictionary contains the metadata for the data. See
        :ref:`dataset_metadata` for more information.

        Further, notice that it is also possible to load a combination of files
        or sets of files each requiring their specific reader. For that
        ``filenames`` needs to be a `dict` (see parameters list below), e.g.::

            scn = Scene(filenames={'nwcsaf-pps_nc': glob('/path/to/nwc/saf/pps/files/*'),
                                   'modis_l1b': glob('/path/to/modis/lvl1/files/*')})


        Args:
            filenames (Iterable or dict): A sequence of files that will be used to load data from. A ``dict`` object
                                          should map reader names to a list of filenames for that reader.
            reader (str or list): The name of the reader to use for loading the data or a list of names.
            filter_parameters (dict): Specify loaded file filtering parameters.
                                      Shortcut for `reader_kwargs['filter_parameters']`.
            reader_kwargs (dict): Keyword arguments to pass to specific reader instances.
                Either a single dictionary that will be passed onto to all
                reader instances, or a dictionary mapping reader names to
                sub-dictionaries to pass different arguments to different
                reader instances.

                Keyword arguments for remote file access are also given in this dictionary.
                See `documentation <https://satpy.readthedocs.io/en/stable/remote_reading.html>`_
                for usage examples.

        """
        self.attrs = dict()

        storage_options, cleaned_reader_kwargs = get_storage_options_from_reader_kwargs(reader_kwargs)

        if filter_parameters:
            if cleaned_reader_kwargs is None:
                cleaned_reader_kwargs = {}
            else:
                cleaned_reader_kwargs = cleaned_reader_kwargs.copy()
            cleaned_reader_kwargs.setdefault("filter_parameters", {}).update(filter_parameters)

        if filenames and isinstance(filenames, str):
            raise ValueError("'filenames' must be a list of files: Scene(filenames=[filename])")

        if filenames:
            filenames = convert_remote_files_to_fsspec(filenames, storage_options)

        self._readers = self._create_reader_instances(filenames=filenames,
                                                      reader=reader,
                                                      reader_kwargs=cleaned_reader_kwargs)
        self._datasets = DatasetDict()
        self._wishlist = set()
        self._dependency_tree = DependencyTree(self._readers)
        self._resamplers = {}

    @property
    def wishlist(self):
        """Return a copy of the wishlist."""
        return self._wishlist.copy()

    def _ipython_key_completions_(self):
        return [x["name"] for x in self._datasets.keys()]

    def _create_reader_instances(self,
                                 filenames=None,
                                 reader=None,
                                 reader_kwargs=None):
        """Find readers and return their instances."""
        return load_readers(filenames=filenames,
                            reader=reader,
                            reader_kwargs=reader_kwargs)

    @property
    def sensor_names(self) -> set[str]:
        """Return sensor names for the data currently contained in this Scene.

        Sensor information is collected from data contained in the Scene
        whether loaded from a reader or generated as a composite with
        :meth:`load` or added manually using ``scn["name"] = data_arr``).
        Sensor information is also collected from any loaded readers.
        In some rare cases this may mean that the reader includes sensor
        information for data that isn't actually loaded or even available.

        """
        contained_sensor_names = self._contained_sensor_names()
        reader_sensor_names = set([sensor for reader_instance in self._readers.values()
                                   for sensor in reader_instance.sensor_names])
        return contained_sensor_names | reader_sensor_names

    def _contained_sensor_names(self) -> set[str]:
        sensor_names = set()
        for data_arr in self.values():
            if "sensor" not in data_arr.attrs:
                continue
            if isinstance(data_arr.attrs["sensor"], str):
                sensor_names.add(data_arr.attrs["sensor"])
            elif isinstance(data_arr.attrs["sensor"], set):
                sensor_names.update(data_arr.attrs["sensor"])
        return sensor_names

    @property
    def start_time(self):
        """Return the start time of the contained data.

        If no data is currently contained in the Scene then loaded readers
        will be consulted.

        """
        start_times = [data_arr.attrs["start_time"] for data_arr in self.values()
                       if "start_time" in data_arr.attrs]
        if not start_times:
            start_times = self._reader_times("start_time")
        if not start_times:
            return None
        return min(start_times)

    @property
    def end_time(self):
        """Return the end time of the file.

        If no data is currently contained in the Scene then loaded readers
        will be consulted. If no readers are loaded then the
        :attr:`Scene.start_time` is returned.

        """
        end_times = [data_arr.attrs["end_time"] for data_arr in self.values()
                     if "end_time" in data_arr.attrs]
        if not end_times:
            end_times = self._reader_times("end_time")
        if not end_times:
            return self.start_time
        return max(end_times)

    def _reader_times(self, time_prop_name):
        return [getattr(reader, time_prop_name) for reader in self._readers.values()]

    @property
    def missing_datasets(self):
        """Set of DataIDs that have not been successfully loaded."""
        return set(self._wishlist) - set(self._datasets.keys())

    def _compare_areas(self, datasets=None, compare_func=max):
        """Compare areas for the provided datasets.

        Args:
            datasets (Iterable): Datasets whose areas will be compared. Can
                                 be either `xarray.DataArray` objects or
                                 identifiers to get the DataArrays from the
                                 current Scene. Defaults to all datasets.
                                 This can also be a series of area objects,
                                 typically AreaDefinitions.
            compare_func (Callable): `min` or `max` or other function used to
                                     compare the dataset's areas.

        """
        if datasets is None:
            datasets = list(self.values())

        areas = self._gather_all_areas(datasets)

        if isinstance(areas[0], AreaDefinition):
            first_crs = areas[0].crs
            if not all(ad.crs == first_crs for ad in areas[1:]):
                raise ValueError("Can't compare areas with different "
                                 "projections.")
            return self._compare_area_defs(compare_func, areas)
        return self._compare_swath_defs(compare_func, areas)

    @staticmethod
    def _compare_area_defs(compare_func: Callable, area_defs: list[AreaDefinition]) -> list[AreaDefinition]:
        def _key_func(area_def: AreaDefinition) -> tuple:
            """Get comparable version of area based on resolution.

            Pixel size x is the primary comparison parameter followed by
            the y dimension pixel size. The extent of the area and the
            name (area_id) of the area are also used to act as
            "tiebreakers" between areas of the same resolution.

            """
            pixel_size_x_inverse = 1. / abs(area_def.pixel_size_x)
            pixel_size_y_inverse = 1. / abs(area_def.pixel_size_y)
            area_id = area_def.area_id
            return pixel_size_x_inverse, pixel_size_y_inverse, area_def.area_extent, area_id
        return compare_func(area_defs, key=_key_func)

    @staticmethod
    def _compare_swath_defs(compare_func: Callable, swath_defs: list[SwathDefinition]) -> list[SwathDefinition]:
        def _key_func(swath_def: SwathDefinition) -> tuple:
            attrs = getattr(swath_def.lons, "attrs", {})
            lon_ds_name = attrs.get("name")
            rev_shape = swath_def.shape[::-1]
            return rev_shape + (lon_ds_name,)
        return compare_func(swath_defs, key=_key_func)

    def _gather_all_areas(self, datasets):
        """Gather all areas from datasets.

        They have to be of the same type, and at least one dataset should have
        an area.
        """
        areas = []
        for ds in datasets:
            if isinstance(ds, BaseDefinition):
                areas.append(ds)
                continue
            elif not isinstance(ds, DataArray):
                ds = self[ds]
            area = ds.attrs.get("area")
            areas.append(area)
        areas = [x for x in areas if x is not None]
        if not areas:
            raise ValueError("No dataset areas available")
        if not all(isinstance(x, type(areas[0]))
                   for x in areas[1:]):
            raise ValueError("Can't compare areas of different types")
        return areas

    def finest_area(self, datasets=None):
        """Get highest resolution area for the provided datasets.

        Args:
            datasets (Iterable): Datasets whose areas will be compared. Can
                                 be either `xarray.DataArray` objects or
                                 identifiers to get the DataArrays from the
                                 current Scene. Defaults to all datasets.

        """
        return self._compare_areas(datasets=datasets, compare_func=max)

    def max_area(self, datasets=None):
        """Get highest resolution area for the provided datasets. Deprecated.

        Deprecated.  Use :meth:`finest_area` instead.

        Args:
            datasets (Iterable): Datasets whose areas will be compared. Can
                                 be either `xarray.DataArray` objects or
                                 identifiers to get the DataArrays from the
                                 current Scene. Defaults to all datasets.

        """
        warnings.warn(
            "'max_area' is deprecated, use 'finest_area' instead.",
            DeprecationWarning,
            stacklevel=2
        )
        return self.finest_area(datasets=datasets)

    def coarsest_area(self, datasets=None):
        """Get lowest resolution area for the provided datasets.

        Args:
            datasets (Iterable): Datasets whose areas will be compared. Can
                                 be either `xarray.DataArray` objects or
                                 identifiers to get the DataArrays from the
                                 current Scene. Defaults to all datasets.

        """
        return self._compare_areas(datasets=datasets, compare_func=min)

    def min_area(self, datasets=None):
        """Get lowest resolution area for the provided datasets. Deprecated.

        Deprecated.  Use :meth:`coarsest_area` instead.

        Args:
            datasets (Iterable): Datasets whose areas will be compared. Can
                                 be either `xarray.DataArray` objects or
                                 identifiers to get the DataArrays from the
                                 current Scene. Defaults to all datasets.

        """
        warnings.warn(
            "'min_area' is deprecated, use 'coarsest_area' instead.",
            DeprecationWarning,
            stacklevel=2
        )
        return self.coarsest_area(datasets=datasets)

    def available_dataset_ids(self, reader_name=None, composites=False):
        """Get DataIDs of loadable datasets.

        This can be for all readers loaded by this Scene or just for
        ``reader_name`` if specified.

        Available dataset names are determined by what each individual reader
        can load. This is normally determined by what files are needed to load
        a dataset and what files have been provided to the scene/reader.
        Some readers dynamically determine what is available based on the
        contents of the files provided.

        By default, only returns non-composite dataset IDs.  To include
        composite dataset IDs, pass ``composites=True``.

        Args:
            reader_name (str, Optional): Name of reader for which to return
                dataset IDs.  If not passed, return dataset IDs for all
                readers.
            composites (bool, Optional): If True, return dataset IDs including
                composites.  If False (default), return only non-composite
                dataset IDs.

        Returns: list of available dataset IDs

        """
        try:
            if reader_name:
                readers = [self._readers[reader_name]]
            else:
                readers = self._readers.values()
        except (AttributeError, KeyError):
            raise KeyError("No reader '%s' found in scene" % reader_name)

        available_datasets = sorted([dataset_id
                                     for reader in readers
                                     for dataset_id in reader.available_dataset_ids])
        if composites:
            available_datasets += sorted(self.available_composite_ids())
        return available_datasets

    def available_dataset_names(self, reader_name=None, composites=False):
        """Get the list of the names of the available datasets.

        By default, this only shows names of datasets directly defined in (one
        of the) readers.  Names of composites are not returned unless the
        argument ``composites=True`` is passed.

        Args:
            reader_name (str, Optional): Name of reader for which to return
                dataset IDs.  If not passed, return dataset names for all
                readers.
            composites (bool, Optional): If True, return dataset IDs including
                composites.  If False (default), return only non-composite
                dataset names.

        Returns: list of available dataset names
        """
        return sorted(set(x["name"] for x in self.available_dataset_ids(
            reader_name=reader_name, composites=composites)))

    def all_dataset_ids(self, reader_name=None, composites=False):
        """Get IDs of all datasets from loaded readers or `reader_name` if specified.

        Excludes composites unless ``composites=True`` is passed.

        Args:
            reader_name (str, Optional): Name of reader for which to return
                dataset IDs.  If not passed, return dataset IDs for all
                readers.
            composites (bool, Optional): If True, return dataset IDs including
                composites.  If False (default), return only non-composite
                dataset IDs.

        Returns: list of all dataset IDs

        """
        try:
            if reader_name:
                readers = [self._readers[reader_name]]
            else:
                readers = self._readers.values()
        except (AttributeError, KeyError):
            raise KeyError("No reader '%s' found in scene" % reader_name)

        all_datasets = [dataset_id
                        for reader in readers
                        for dataset_id in reader.all_dataset_ids]
        if composites:
            all_datasets += self.all_composite_ids()
        return all_datasets

    def all_dataset_names(self, reader_name=None, composites=False):
        """Get all known dataset names configured for the loaded readers.

        Note that some readers dynamically determine what datasets are known
        by reading the contents of the files they are provided. This means
        that the list of datasets returned by this method may change depending
        on what files are provided even if a product/dataset is a "standard"
        product for a particular reader.

        Excludes composites unless ``composites=True`` is passed.

        Args:
            reader_name (str, Optional): Name of reader for which to return
                dataset IDs.  If not passed, return dataset names for all
                readers.
            composites (bool, Optional): If True, return dataset IDs including
                composites.  If False (default), return only non-composite
                dataset names.

        Returns: list of all dataset names

        """
        return sorted(set(x["name"] for x in self.all_dataset_ids(
            reader_name=reader_name, composites=composites)))

    def _check_known_composites(self, available_only=False):
        """Create new dependency tree and check what composites we know about."""
        # Note if we get compositors from the dep tree then it will include
        # modified composites which we don't want
        sensor_comps, mods = load_compositor_configs_for_sensors(self.sensor_names)
        # recreate the dependency tree so it doesn't interfere with the user's
        # wishlist from self._dependency_tree
        dep_tree = DependencyTree(self._readers, sensor_comps, mods, available_only=available_only)
        # ignore inline compositor dependencies starting with '_'
        comps = (comp for comp_dict in sensor_comps.values()
                 for comp in comp_dict.keys() if not comp["name"].startswith("_"))
        # make sure that these composites are even create-able by these readers
        all_comps = set(comps)
        # find_dependencies will update the all_comps set with DataIDs
        try:
            dep_tree.populate_with_keys(all_comps)
        except MissingDependencies:
            pass
        available_comps = set(x.name for x in dep_tree.trunk())
        # get rid of modified composites that are in the trunk
        return sorted(available_comps & all_comps)

    def available_composite_ids(self):
        """Get IDs of composites that can be generated from the available datasets."""
        return self._check_known_composites(available_only=True)

    def available_composite_names(self):
        """Names of all configured composites known to this Scene."""
        return sorted(set(x["name"] for x in self.available_composite_ids()))

    def all_composite_ids(self):
        """Get all IDs for configured composites."""
        return self._check_known_composites()

    def all_composite_names(self):
        """Get all names for all configured composites."""
        return sorted(set(x["name"] for x in self.all_composite_ids()))

    def all_modifier_names(self):
        """Get names of configured modifier objects."""
        return sorted(self._dependency_tree.modifiers.keys())

    def __str__(self):
        """Generate a nice print out for the scene."""
        res = (str(proj) for proj in self._datasets.values())
        return "\n".join(res)

    def __iter__(self):
        """Iterate over the datasets."""
        for x in self._datasets.values():
            yield x

    def iter_by_area(self):
        """Generate datasets grouped by Area.

        :return: generator of (area_obj, list of dataset objects)
        """
        datasets_by_area = {}
        for ds in self:
            a = ds.attrs.get("area")
            dsid = DataID.from_dataarray(ds)
            datasets_by_area.setdefault(a, []).append(dsid)

        return datasets_by_area.items()

    def keys(self, **kwargs):
        """Get DataID keys for the underlying data container."""
        return self._datasets.keys(**kwargs)

    def values(self):
        """Get values for the underlying data container."""
        return self._datasets.values()

    def _copy_datasets_and_wishlist(self, new_scn, datasets):
        for ds_id in datasets:
            # NOTE: Must use `._datasets` or side effects of `__setitem__`
            #       could hurt us with regards to the wishlist
            new_scn._datasets[ds_id] = self[ds_id]
        new_scn._wishlist = self._wishlist.copy()

    def copy(self, datasets=None):
        """Create a copy of the Scene including dependency information.

        Args:
            datasets (list, tuple): `DataID` objects for the datasets
                                    to include in the new Scene object.

        """
        new_scn = self.__class__()
        new_scn.attrs = self.attrs.copy()
        new_scn._dependency_tree = self._dependency_tree.copy()
        if datasets is None:
            datasets = self.keys()
        self._copy_datasets_and_wishlist(new_scn, datasets)
        return new_scn

    @property
    def all_same_area(self):
        """All contained data arrays are on the same area."""
        all_areas = [x.attrs.get("area", None) for x in self.values()]
        all_areas = [x for x in all_areas if x is not None]
        return all(all_areas[0] == x for x in all_areas[1:])

    @property
    def all_same_proj(self):
        """All contained data array are in the same projection."""
        all_areas = [x.attrs.get("area", None) for x in self.values()]
        all_areas = [x for x in all_areas if x is not None]
        return all(all_areas[0].crs == x.crs for x in all_areas[1:])

    @staticmethod
    def _slice_area_from_bbox(src_area, dst_area, ll_bbox=None,
                              xy_bbox=None):
        """Slice the provided area using the bounds provided."""
        if ll_bbox is not None:
            dst_area = AreaDefinition(
                "crop_area", "crop_area", "crop_latlong",
                {"proj": "latlong"}, 100, 100, ll_bbox)
        elif xy_bbox is not None:
            dst_area = AreaDefinition(
                "crop_area", "crop_area", "crop_xy",
                src_area.crs, src_area.width, src_area.height,
                xy_bbox)
        x_slice, y_slice = src_area.get_area_slices(dst_area)
        return src_area[y_slice, x_slice], y_slice, x_slice

    def _slice_datasets(self, dataset_ids, slice_key, new_area, area_only=True):
        """Slice scene in-place for the datasets specified."""
        new_datasets = {}
        datasets = (self[ds_id] for ds_id in dataset_ids)
        for ds, parent_ds in dataset_walker(datasets):
            ds_id = DataID.from_dataarray(ds)
            # handle ancillary variables
            pres = None
            if parent_ds is not None:
                parent_dsid = DataID.from_dataarray(parent_ds)
                pres = new_datasets[parent_dsid]
            if ds_id in new_datasets:
                replace_anc(ds, pres)
                continue
            if area_only and ds.attrs.get("area") is None:
                new_datasets[ds_id] = ds
                replace_anc(ds, pres)
                continue

            if not isinstance(slice_key, dict):
                # match dimension name to slice object
                key = dict(zip(ds.dims, slice_key))
            else:
                key = slice_key
            new_ds = ds.isel(**key)
            if new_area is not None:
                new_ds.attrs["area"] = new_area

            new_datasets[ds_id] = new_ds
            if parent_ds is None:
                # don't use `__setitem__` because we don't want this to
                # affect the existing wishlist/dep tree
                self._datasets[ds_id] = new_ds
            else:
                replace_anc(new_ds, pres)

    def slice(self, key):  # noqa: A003
        """Slice Scene by dataset index.

        .. note::

            DataArrays that do not have an ``area`` attribute will not be
            sliced.

        """
        if not self.all_same_area:
            raise RuntimeError("'Scene' has different areas and cannot "
                               "be usefully sliced.")
        # slice
        new_scn = self.copy()
        new_scn._wishlist = self._wishlist
        for area, dataset_ids in self.iter_by_area():
            if area is not None:
                # assume dimensions for area are y and x
                one_ds = self[dataset_ids[0]]
                area_key = tuple(sl for dim, sl in zip(one_ds.dims, key) if dim in ["y", "x"])
                new_area = area[area_key]
            else:
                new_area = None
            new_scn._slice_datasets(dataset_ids, key, new_area)
        return new_scn

    def crop(
            self,
            area: AreaDefinition | None = None,
            ll_bbox: tuple[float, float, float, float] | None = None,
            xy_bbox: tuple[float, float, float, float] | None = None,
            dataset_ids: Iterable | None = None,
    ) -> Scene:
        """Crop Scene to a specific Area boundary or bounding box.

        Args:
            area: Area to crop the current Scene to
            ll_bbox: 4-element tuple where values are in
                lon/lat degrees. Elements are
                ``(xmin, ymin, xmax, ymax)`` where X is
                longitude and Y is latitude.
            xy_bbox: Same as `ll_bbox` but elements are in projection units.
            dataset_ids: DataIDs to include in the returned `Scene`. Defaults to all datasets.

        This method will attempt to intelligently slice the data to preserve
        relationships between datasets. For example, if we are cropping two
        DataArrays of 500m and 1000m pixel resolution then this method will
        assume that exactly 4 pixels of the 500m array cover the same
        geographic area as a single 1000m pixel. It handles these cases based
        on the shapes of the input arrays and adjusting slicing indexes
        accordingly. This method will have trouble handling cases where data
        arrays seem related but don't cover the same geographic area or if the
        coarsest resolution data is not related to the other arrays which are
        related.

        It can be useful to follow cropping with a call to the native
        resampler to resolve all datasets to the same resolution and compute
        any composites that could not be generated previously::

        >>> cropped_scn = scn.crop(ll_bbox=(-105., 40., -95., 50.))
        >>> remapped_scn = cropped_scn.resample(resampler='native')

        .. note::

            The `resample` method automatically crops input data before
            resampling to save time/memory.

        """
        if len([x for x in [area, ll_bbox, xy_bbox] if x is not None]) != 1:
            raise ValueError("One and only one of 'area', 'll_bbox', "
                             "or 'xy_bbox' can be specified.")

        new_scn = self.copy(datasets=dataset_ids)
        if not new_scn.all_same_proj and xy_bbox is not None:
            raise ValueError("Can't crop when dataset_ids are not all on the "
                             "same projection.")

        # get the lowest resolution area, use it as the base of the slice
        # this makes sure that the other areas *should* be a consistent factor
        coarsest_area = new_scn.coarsest_area()
        if isinstance(area, str):
            area = get_area_def(area)
        new_coarsest_area, min_y_slice, min_x_slice = self._slice_area_from_bbox(
            coarsest_area, area, ll_bbox, xy_bbox)
        new_target_areas = {}
        for src_area, ids_on_area in new_scn.iter_by_area():
            if src_area is None:
                for ds_id in ids_on_area:
                    new_scn._datasets[ds_id] = self[ds_id]
                continue

            y_factor, y_remainder = np.divmod(float(src_area.shape[0]),
                                              coarsest_area.shape[0])
            x_factor, x_remainder = np.divmod(float(src_area.shape[1]),
                                              coarsest_area.shape[1])
            y_factor = int(y_factor)
            x_factor = int(x_factor)
            if y_remainder == 0 and x_remainder == 0:
                y_slice = slice(min_y_slice.start * y_factor,
                                min_y_slice.stop * y_factor)
                x_slice = slice(min_x_slice.start * x_factor,
                                min_x_slice.stop * x_factor)
                new_area = src_area[y_slice, x_slice]
                slice_key = {"y": y_slice, "x": x_slice}
                new_scn._slice_datasets(ids_on_area, slice_key, new_area)
            else:
                new_target_areas[src_area] = self._slice_area_from_bbox(
                    src_area, area, ll_bbox, xy_bbox
                )

        return new_scn

    def aggregate(self, dataset_ids=None, boundary="trim", side="left", func="mean", **dim_kwargs):
        """Create an aggregated version of the Scene.

        Args:
            dataset_ids (Iterable): DataIDs to include in the returned
                                    `Scene`. Defaults to all datasets.
            func (str, Callable): Function to apply on each aggregation window. One of
                           'mean', 'sum', 'min', 'max', 'median', 'argmin',
                           'argmax', 'prod', 'std', 'var' strings or a custom
                           function. 'mean' is the default.
            boundary: See :meth:`xarray.DataArray.coarsen`, 'trim' by default.
            side: See :meth:`xarray.DataArray.coarsen`, 'left' by default.
            dim_kwargs: the size of the windows to aggregate.

        Returns:
            A new aggregated scene

        See Also:
            xarray.DataArray.coarsen

        Example:
            `scn.aggregate(func='min', x=2, y=2)` will apply the `min` function
            across a window of size 2 pixels.

        """
        new_scn = self.copy(datasets=dataset_ids)

        for src_area, ds_ids in new_scn.iter_by_area():
            if src_area is None:
                for ds_id in ds_ids:
                    new_scn._datasets[ds_id] = self[ds_id]
                continue

            target_area = src_area.aggregate(boundary=boundary, **dim_kwargs)
            resolution = _get_area_resolution(target_area)
            for ds_id in ds_ids:
                new_scn._datasets[ds_id] = _aggregate_data_array(self[ds_id],
                                                                 func=func,
                                                                 boundary=boundary,
                                                                 side=side,
                                                                 **dim_kwargs)
                new_scn._datasets[ds_id].attrs = self[ds_id].attrs.copy()
                new_scn._datasets[ds_id].attrs["area"] = target_area
                new_scn._datasets[ds_id].attrs["resolution"] = resolution
        return new_scn

    def get(self, key, default=None):
        """Return value from DatasetDict with optional default."""
        return self._datasets.get(key, default)

    def __getitem__(self, key):
        """Get a dataset or create a new 'slice' of the Scene."""
        if isinstance(key, tuple):
            return self.slice(key)
        return self._datasets[key]

    def __setitem__(self, key, value):
        """Add the item to the scene."""
        self._datasets[key] = value
        # this could raise a KeyError but never should in this case
        ds_id = self._datasets.get_key(key)
        self._wishlist.add(ds_id)
        self._dependency_tree.add_leaf(ds_id)

    def __delitem__(self, key):
        """Remove the item from the scene."""
        k = self._datasets.get_key(key)
        self._wishlist.discard(k)
        del self._datasets[k]

    def __contains__(self, name):
        """Check if the dataset is in the scene."""
        return name in self._datasets

    def _slice_data(self, source_area, slices, dataset):
        """Slice the data to reduce it."""
        slice_x, slice_y = slices
        dataset = dataset.isel(x=slice_x, y=slice_y)
        if ("x", source_area.width) not in dataset.sizes.items():
            raise RuntimeError
        if ("y", source_area.height) not in dataset.sizes.items():
            raise RuntimeError
        dataset.attrs["area"] = source_area

        return dataset

    def _resampled_scene(self, new_scn, destination_area, reduce_data=True,
                         **resample_kwargs):
        """Resample `datasets` to the `destination` area.

        If data reduction is enabled, some local caching is perfomed in order to
        avoid recomputation of area intersections.
        """
        from satpy.resample.base import resample_dataset

        new_datasets = {}
        datasets = list(new_scn._datasets.values())
        destination_area = self._get_finalized_destination_area(destination_area, new_scn)

        resamplers = {}
        reductions = {}
        for dataset, parent_dataset in dataset_walker(datasets):
            ds_id = DataID.from_dataarray(dataset)
            pres = None
            if parent_dataset is not None:
                pres = new_datasets[DataID.from_dataarray(parent_dataset)]
            if ds_id in new_datasets:
                replace_anc(new_datasets[ds_id], pres)
                if ds_id in new_scn._datasets:
                    new_scn._datasets[ds_id] = new_datasets[ds_id]
                continue
            if dataset.attrs.get("area") is None:
                if parent_dataset is None:
                    new_scn._datasets[ds_id] = dataset
                else:
                    replace_anc(dataset, pres)
                continue
            LOG.debug("Resampling %s", ds_id)
            source_area = dataset.attrs["area"]
            dataset, source_area = self._reduce_data(dataset, source_area, destination_area,
                                                     reduce_data, reductions, resample_kwargs)
            self._prepare_resampler(source_area, destination_area, resamplers, resample_kwargs)
            kwargs = resample_kwargs.copy()
            kwargs["resampler"] = resamplers[source_area]
            res = resample_dataset(dataset, destination_area, **kwargs)
            new_datasets[ds_id] = res
            if ds_id in new_scn._datasets:
                new_scn._datasets[ds_id] = res
            if parent_dataset is not None:
                replace_anc(res, pres)

    def _get_finalized_destination_area(self, destination_area, new_scn):
        if isinstance(destination_area, str):
            destination_area = get_area_def(destination_area)
        if hasattr(destination_area, "freeze"):
            try:
                finest_area = new_scn.finest_area()
                destination_area = destination_area.freeze(finest_area)
            except ValueError:
                raise ValueError("No dataset areas available to freeze "
                                 "DynamicAreaDefinition.")
        return destination_area

    def _prepare_resampler(self, source_area, destination_area, resamplers, resample_kwargs):
        from satpy.resample.base import prepare_resampler

        if source_area not in resamplers:
            key, resampler = prepare_resampler(
                source_area, destination_area, **resample_kwargs)
            resamplers[source_area] = resampler
            self._resamplers[key] = resampler

    def _reduce_data(self, dataset, source_area, destination_area, reduce_data, reductions, resample_kwargs):
        try:
            if reduce_data:
                key = source_area
                try:
                    (slice_x, slice_y), source_area = reductions[key]
                except KeyError:
                    if resample_kwargs.get("resampler") == "gradient_search":
                        factor = resample_kwargs.get("shape_divisible_by", 2)
                    else:
                        factor = None
                    try:
                        slice_x, slice_y = source_area.get_area_slices(
                            destination_area, shape_divisible_by=factor)
                    except TypeError:
                        slice_x, slice_y = source_area.get_area_slices(
                            destination_area)
                    source_area = source_area[slice_y, slice_x]
                    reductions[key] = (slice_x, slice_y), source_area
                dataset = self._slice_data(source_area, (slice_x, slice_y), dataset)
            else:
                LOG.debug("Data reduction disabled by the user")
        except NotImplementedError:
            LOG.info("Not reducing data before resampling.")
        return dataset, source_area

    def resample(
            self,
            destination: AreaDefinition | CoordinateDefinition | str | None = None,
            datasets: Iterable | None = None,
            generate: bool = True,
            unload: bool = True,
            resampler: str | None = None,
            reduce_data: bool = True,
            **resample_kwargs,
    ) -> Scene:
        """Resample datasets and return a new scene.

        Args:
            destination: area definition to
                resample to. If not specified, then the area returned by
                `Scene.finest_area()` will be used. Inputs can be
                pyresample geometry objects, or
                a string containing the name of an area
                defined in the ``areas.yaml`` file. See :doc:`/resample`
                for the list of pre-configured areas already available in satpy
                and :ref:`component_configuration`
                on how to configure your own ``areas.yaml`` file.
            datasets: Limit datasets to resample to these specified
                data arrays. By default all currently loaded
                datasets are resampled.
            generate: Generate any requested composites that could not
                be previously due to incompatible areas (default: True).
            unload: Remove any datasets no longer needed after
                requested composites have been generated (default: True).
            resampler: Name of resampling method to use. By default,
                this is a nearest neighbor KDTree-based resampling
                ('nearest'). Other possible values include 'native', 'ewa',
                etc. See the :mod:`~satpy.resample` documentation for more
                information.
            reduce_data: Reduce data by matching the input and output
                areas and slicing the data arrays (default: True)
            resample_kwargs: Remaining keyword arguments to pass to individual
                resampler classes. See the individual resampler class
                documentation :mod:`here <satpy.resample>` for available
                arguments.

        """
        if destination is None:
            destination = self.finest_area(datasets)
        new_scn = self.copy(datasets=datasets)
        self._resampled_scene(new_scn, destination, resampler=resampler,
                              reduce_data=reduce_data, **resample_kwargs)

        # regenerate anything from the wishlist that needs it (combining
        # multiple resolutions, etc.)
        if generate:
            new_scn.generate_possible_composites(unload)

        return new_scn

    def show(self, dataset_id, overlay=None):
        """Show the *dataset* on screen as an image.

        Show dataset on screen as an image, possibly with an overlay.

        Args:
            dataset_id (DataID, DataQuery or str):
                Either a DataID, a DataQuery or a string, that refers to a data
                array that has been previously loaded using Scene.load.
            overlay (dict, Optional):
                Add an overlay before showing the image.  The keys/values for
                this dictionary are as the arguments for
                :meth:`~satpy.enhancements.overlays.add_overlay`.  The dictionary should
                contain at least the key ``"coast_dir"``, which should refer
                to a top-level directory containing shapefiles.  See the
                pycoast_ package documentation for coastline shapefile
                installation instructions.

        .. _pycoast: https://pycoast.readthedocs.io/

        """
        from satpy.enhancements.enhancer import get_enhanced_image
        from satpy.utils import in_ipynb
        img = get_enhanced_image(self[dataset_id].squeeze(), overlay=overlay)
        if not in_ipynb():
            img.show()
        return img

    def to_geoviews(
            self,
            gvtype: Any | None = None,
            datasets: list | None = None,
            vdims: list[str] | None = None,
            dynamic: bool = False,
    ):
        """Convert satpy Scene to geoviews.

        Args:
            scn: Satpy Scene.
            gvtype:
                One of gv.Image, gv.LineContours, gv.FilledContours, gv.Points
                Default to :class:`geoviews.Image <geoviews.element.geo.Image>`.
                See Geoviews documentation for details.
            datasets: Limit included products to these datasets
            vdims:
                Value dimensions. See geoviews documentation for more information.
                If not given defaults to first data variable
            dynamic: Load and compute data on-the-fly during
                visualization. Default is ``False``. See
                https://holoviews.org/user_guide/Gridded_Datasets.html#working-with-xarray-data-types
                for more information. Has no effect when data to be visualized
                only has 2 dimensions (y/x or longitude/latitude) and doesn't
                require grouping via the Holoviews ``groupby`` function.

        Returns: geoviews object

        Todo:
            * better handling of projection information in datasets which are
              to be passed to geoviews

        """
        from satpy._scene_converters import to_geoviews
        return to_geoviews(self, gvtype=gvtype, datasets=datasets,
                           vdims=vdims, dynamic=dynamic)


    def to_hvplot(self, datasets=None, *args, **kwargs):
        """Convert satpy Scene to Hvplot. The method could not be used with composites of swath data.

        Args:
            scn (satpy.scene.Scene): Satpy Scene.
            datasets (list): Limit included products to these datasets.
            args: Arguments coming from hvplot
            kwargs: hvplot options dictionary.

        Returns:
            hvplot object that contains within it the plots of datasets list.
            As default it contains all Scene datasets plots and a plot title
            is shown.

        Example usage::

           scene_list = ['ash','IR_108']
           scn = Scene()
           scn.load(scene_list)
           scn = scn.resample('eurol')
           plot = scn.to_hvplot(datasets=scene_list)
           plot.ash+plot.IR_108
        """
        from satpy._scene_converters import to_hvplot

        return to_hvplot(self, datasets=None, *args, **kwargs)



    def to_xarray_dataset(self, datasets=None, compat="minimal"):
        """Merge all xr.DataArrays of a scene to a xr.DataSet.

        Parameters:
            datasets (list):
                List of products to include in the :class:`xarray.Dataset`
            compat (str):
                How to compare variables with the same name for conflicts.
                See :func:`xarray.merge` for possible options. Defaults to
                "minimal" which drops conflicting variables.

        Returns: :class:`xarray.Dataset`

        """
        from satpy._scene_converters import _get_dataarrays_from_identifiers

        dataarrays = _get_dataarrays_from_identifiers(self, datasets)

        if len(dataarrays) == 0:
            return xr.Dataset()

        ds_dict = {i.attrs["name"]: i.rename(i.attrs["name"]) for i in dataarrays if i.attrs.get("area") is not None}
        mdata = combine_metadata(*tuple(i.attrs for i in dataarrays))
        if mdata.get("area") is None or not isinstance(mdata["area"], SwathDefinition):
            # either don't know what the area is or we have an AreaDefinition
            ds = xr.merge(ds_dict.values(), compat=compat)
        else:
            # we have a swath definition and should use lon/lat values
            lons, lats = mdata["area"].get_lonlats()
            if not isinstance(lons, DataArray):
                lons = DataArray(lons, dims=("y", "x"))
                lats = DataArray(lats, dims=("y", "x"))
            ds = xr.Dataset(ds_dict, coords={"latitude": lats,
                                             "longitude": lons})

        ds.attrs = mdata
        return ds

    def to_xarray(
            self,
            datasets: Iterable | None = None,
            header_attrs: dict | None = None,
            exclude_attrs: Iterable | None = None,
            flatten_attrs: bool = False,
            pretty: bool = True,
            include_lonlats: bool = True,
            epoch: str | None = None,
            include_orig_name: bool = True,
            numeric_name_prefix: str = "CHANNEL_",
    ) -> xr.Dataset:
        """Merge all xr.DataArray(s) of a satpy.Scene to a CF-compliant xarray object.

        If all Scene DataArrays are on the same area, it returns an xr.Dataset.
        If Scene DataArrays are on different areas, currently it fails, although
        in future we might return a DataTree object, grouped by area.

        Args:
            datasets:
                List of Satpy Scene datasets to include in the output xr.Dataset.
                Elements can be string name, a wavelength as a number, a DataID,
                or DataQuery object.
                If None (the default), it include all loaded Scene datasets.
            header_attrs:
                Global attributes of the output xr.Dataset.
            exclude_attrs:
                List of xr.DataArray attribute names to be excluded.
            flatten_attrs:
                If True, flatten dict-type attributes.
            pretty:
                Don't modify coordinate names, if possible. Makes the file prettier,
                but possibly less consistent.
            include_lonlats:
                If True, it includes 'latitude' and 'longitude' coordinates.
                If the 'area' attribute is a SwathDefinition, it always includes
                latitude and longitude coordinates.
            epoch:
                Reference time for encoding the time coordinates (if available).
                Example format: "seconds since 1970-01-01 00:00:00".
                If None, the default reference time is defined using "from satpy.cf.coords import EPOCH"
            include_orig_name:
                Include the original dataset name as a variable attribute in the xr.Dataset.
            numeric_name_prefix:
                Prefix to add the each variable with name starting with a digit.
                Use '' or None to leave this out.

        Returns:
            A CF-compliant xr.Dataset

        """
        from satpy._scene_converters import to_xarray

        return to_xarray(scn=self,
                         datasets=datasets,  # DataID
                         header_attrs=header_attrs,
                         exclude_attrs=exclude_attrs,
                         flatten_attrs=flatten_attrs,
                         pretty=pretty,
                         include_lonlats=include_lonlats,
                         epoch=epoch,
                         include_orig_name=include_orig_name,
                         numeric_name_prefix=numeric_name_prefix)

    def save_dataset(self, dataset_id, filename=None, writer=None,
                     overlay=None, decorate=None, compute=True, **kwargs):
        """Save the ``dataset_id`` to file using ``writer``.

        Args:
            dataset_id (str or numbers.Number or DataID or DataQuery): Identifier for
                the dataset to save to disk.
            filename (str): Optionally specify the filename to save this
                            dataset to. It may include string formatting
                            patterns that will be filled in by dataset
                            attributes.
            writer (str): Name of writer to use when writing data to disk.
                Default to ``"geotiff"``. If not provided, but ``filename`` is
                provided then the filename's extension is used to determine
                the best writer to use.
            overlay (dict): See :func:`satpy.enhancements.overlays.add_overlay`. Only valid
                for "image" writers like `geotiff` or `simple_image`.
            decorate (dict): See :func:`satpy.enhancements.overlays.add_decorate`. Only valid
                for "image" writers like `geotiff` or `simple_image`.
            compute (bool): If `True` (default), compute all of the saves to
                disk. If `False` then the return value is either a
                :doc:`dask:delayed` object or two lists to be passed to
                a `dask.array.store` call. See return values below for more
                details.
            kwargs: Additional writer arguments. See :doc:`../writing` for more
                information.

        Returns:
            Value returned depends on `compute`. If `compute` is `True` then
            the return value is the result of computing a
            :doc:`dask:delayed` object or running :func:`dask.array.store`.
            If `compute` is `False` then the returned value is either a
            :doc:`dask:delayed` object that can be computed using
            `delayed.compute()` or a tuple of (source, target) that should be
            passed to :func:`dask.array.store`. If target is provided the the
            caller is responsible for calling `target.close()` if the target
            has this method.

        """
        from satpy.writers.core.config import load_writer

        if writer is None and filename is None:
            writer = "geotiff"
        elif writer is None:
            writer = self._get_writer_by_ext(os.path.splitext(filename)[1])

        writer, save_kwargs = load_writer(writer,
                                          filename=filename,
                                          **kwargs)
        return writer.save_dataset(self[dataset_id],
                                   overlay=overlay, decorate=decorate,
                                   compute=compute, **save_kwargs)

    def save_datasets(self, writer=None, filename=None, datasets=None, compute=True,
                      **kwargs):
        """Save requested datasets present in a scene to disk using ``writer``.

        Note that dependency datasets (those loaded solely to create another
        and not requested explicitly) that may be contained in this Scene will
        not be saved by default. The default datasets are those explicitly
        requested through ``.load`` and exist in the Scene currently. Specify
        dependency datasets using the ``datasets`` keyword argument.

        Args:
            writer (str): Name of writer to use when writing data to disk.
                Default to ``"geotiff"``. If not provided, but ``filename`` is
                provided then the filename's extension is used to determine
                the best writer to use.
            filename (str): Optionally specify the filename to save this
                            dataset to. It may include string formatting
                            patterns that will be filled in by dataset
                            attributes.
            datasets (Iterable): Limit written products to these datasets.
                Elements can be string name, a wavelength as a number, a
                DataID, or DataQuery object.
            compute (bool): If `True` (default), compute all of the saves to
                disk. If `False` then the return value is either a
                :doc:`dask:delayed` object or two lists to be passed to
                a `dask.array.store` call. See return values below for more
                details.
            kwargs: Additional writer arguments. See :doc:`../writing` for more
                information.

        Returns:
            Value returned depends on `compute` keyword argument. If
            `compute` is `True` the value is the result of a either a
            `dask.array.store` operation or a :doc:`dask:delayed`
            compute, typically this is `None`. If `compute` is `False` then the
            result is either a :doc:`dask:delayed` object that can be
            computed with `delayed.compute()` or a two element tuple of
            sources and targets to be passed to :func:`dask.array.store`. If
            `targets` is provided then it is the caller's responsibility to
            close any objects that have a "close" method. It is recommended to
            use the utility function
            :func:`~satpy.writers.core.compute.compute_writer_results`.

        """
        from satpy._scene_converters import _get_dataarrays_from_identifiers
        from satpy.writers.core.config import load_writer

        dataarrays = _get_dataarrays_from_identifiers(self, datasets)
        if not dataarrays:
            raise RuntimeError("None of the requested datasets have been "
                               "generated or could not be loaded. Requested "
                               "composite inputs may need to have matching "
                               "dimensions (eg. through resampling).")
        if writer is None:
            if filename is None:
                writer = "geotiff"
            else:
                writer = self._get_writer_by_ext(os.path.splitext(filename)[1])
        writer, save_kwargs = load_writer(writer,
                                          filename=filename,
                                          **kwargs)
        return writer.save_datasets(dataarrays, compute=compute, **save_kwargs)

    def compute(self, **kwargs):
        """Call `compute` on all Scene data arrays.

        See :meth:`xarray.DataArray.compute` for more details.
        Note that this will convert the contents of the DataArray to numpy arrays which
        may not work with all parts of Satpy which may expect dask arrays.
        """
        from dask import compute
        new_scn = self.copy()
        datasets = compute(*(new_scn._datasets.values()), **kwargs)

        for i, k in enumerate(new_scn._datasets.keys()):
            new_scn[k] = datasets[i]

        return new_scn

    def persist(self, **kwargs):
        """Call `persist` on all Scene data arrays.

        See :meth:`xarray.DataArray.persist` for more details.
        """
        from dask import persist
        new_scn = self.copy()
        datasets = persist(*(new_scn._datasets.values()), **kwargs)

        for i, k in enumerate(new_scn._datasets.keys()):
            new_scn[k] = datasets[i]

        return new_scn

    def chunk(self, **kwargs):
        """Call `chunk` on all Scene  data arrays.

        See :meth:`xarray.DataArray.chunk` for more details.
        """
        new_scn = self.copy()
        for k in new_scn._datasets.keys():
            new_scn[k] = new_scn[k].chunk(**kwargs)

        return new_scn

    @staticmethod
    def _get_writer_by_ext(extension):
        """Find the writer matching the ``extension``.

        Defaults to "simple_image".

        Example Mapping:

            - geotiff: .tif, .tiff
            - cf: .nc
            - mitiff: .mitiff
            - simple_image: .png, .jpeg, .jpg, ...

        Args:
            extension (str): Filename extension starting with
                "." (ex. ".png").

        Returns:
            str: The name of the writer to use for this extension.

        """
        mapping = {".tiff": "geotiff", ".tif": "geotiff", ".nc": "cf",
                   ".mitiff": "mitiff"}
        return mapping.get(extension.lower(), "simple_image")

    def _remove_failed_datasets(self, keepables):
        """Remove the datasets that we couldn't create."""
        # copy the set of missing datasets because they won't be valid
        # after they are removed in the next line
        missing = self.missing_datasets.copy()

        keepables = keepables or set()
        # remove reader datasets that couldn't be loaded so they aren't
        # attempted again later
        for n in self.missing_datasets:
            if n not in keepables:
                self._wishlist.discard(n)

        missing_str = ", ".join(str(x) for x in missing)
        LOG.warning("The following datasets were not created and may require "
                    "resampling to be generated: {}".format(missing_str))

    def unload(self, keepables=None):
        """Unload all unneeded datasets.

        Datasets are considered unneeded if they weren't directly requested
        or added to the Scene by the user or they are no longer needed to
        generate composites that have yet to be generated.

        Args:
            keepables (Iterable): DataIDs to keep whether they are needed
                                  or not.

        """
        to_del = [ds_id for ds_id, projectable in self._datasets.items()
                  if ds_id not in self._wishlist and (not keepables or ds_id
                                                      not in keepables)]
        for ds_id in to_del:
            LOG.debug("Unloading dataset: %r", ds_id)
            del self._datasets[ds_id]

    def load(self, wishlist, calibration="*", resolution="*",  # noqa: D417
             polarization="*", level="*", modifiers="*", generate=True, unload=True,
             **kwargs):
        """Read and generate requested datasets.

        When the `wishlist` contains `DataQuery` objects they can either be
        fully-specified `DataQuery` objects with every parameter specified
        or they can not provide certain parameters and the "best" parameter
        will be chosen. For example, if a dataset is available in multiple
        resolutions and no resolution is specified in the wishlist's DataQuery
        then the highest (the smallest number) resolution will be chosen.

        Loaded `DataArray` objects are created and stored in the Scene object.

        Args:
            wishlist (Iterable): List of names (str), wavelengths (float),
                DataQuery objects or DataID of the requested datasets to load.
                See `available_dataset_ids()` for what datasets are available.
            calibration (list | str): Calibration levels to limit available
                datasets. This is a shortcut to having to list each
                DataQuery/DataID in `wishlist`.
            resolution (list | float): Resolution to limit available datasets.
                This is a shortcut similar to calibration.
            polarization (list | str): Polarization ('V', 'H') to limit
                available datasets. This is a shortcut similar to calibration.
            modifiers (tuple | str): Modifiers that should be applied to the
                loaded datasets. This is a shortcut similar to calibration,
                but only represents a single set of modifiers as a tuple.
                For example, specifying
                ``modifiers=('sunz_corrected', 'rayleigh_corrected')`` will
                attempt to apply both of these modifiers to all loaded
                datasets in the specified order ('sunz_corrected' first).
            level (list | str): Pressure level to limit available datasets.
                Pressure should be in hPa or mb. If an altitude is used it
                should be specified in inverse meters (1/m). The units of this
                parameter ultimately depend on the reader.
            generate (bool): Generate composites from the loaded datasets
                (default: True)
            unload (bool): Unload datasets that were required to generate the
                requested datasets (composite dependencies) but are no longer
                needed.

        """
        if isinstance(wishlist, str):
            raise TypeError("'load' expects a list of datasets, got a string.")
        dataset_keys = set(wishlist)
        needed_datasets = (self._wishlist | dataset_keys) - set(self._datasets.keys())
        query = DataQuery(calibration=calibration,
                          polarization=polarization,
                          resolution=resolution,
                          modifiers=modifiers,
                          level=level)
        self._update_dependency_tree(needed_datasets, query)

        self._wishlist |= needed_datasets

        self._read_datasets_from_storage(**kwargs)
        if generate:
            self.generate_possible_composites(unload)

    def _update_dependency_tree(self, needed_datasets, query):
        try:
            comps, mods = load_compositor_configs_for_sensors(self.sensor_names)
            self._dependency_tree.update_compositors_and_modifiers(comps, mods)
            self._dependency_tree.populate_with_keys(needed_datasets, query)
        except MissingDependencies as err:
            raise KeyError(str(err))

    def _read_datasets_from_storage(self, **kwargs):
        """Load datasets from the necessary reader.

        Args:
            **kwargs: Keyword arguments to pass to the reader's `load` method.

        Returns:
            DatasetDict of loaded datasets

        """
        nodes = self._dependency_tree.leaves(limit_nodes_to=self.missing_datasets)
        return self._read_dataset_nodes_from_storage(nodes, **kwargs)

    def _read_dataset_nodes_from_storage(self, reader_nodes, **kwargs):
        """Read the given dataset nodes from storage."""
        # Sort requested datasets by reader
        reader_datasets = self._sort_dataset_nodes_by_reader(reader_nodes)
        loaded_datasets = self._load_datasets_by_readers(reader_datasets, **kwargs)
        self._datasets.update(loaded_datasets)
        return loaded_datasets

    def _sort_dataset_nodes_by_reader(self, reader_nodes):
        reader_datasets = {}
        for node in reader_nodes:
            ds_id = node.name
            # if we already have this node loaded or the node was assigned
            # by the user (node data is None) then don't try to load from a
            # reader
            if ds_id in self._datasets or not isinstance(node, ReaderNode):
                continue
            reader_name = node.reader_name
            if reader_name is None:
                # This shouldn't be possible
                raise RuntimeError("Dependency tree has a corrupt node.")
            reader_datasets.setdefault(reader_name, set()).add(ds_id)
        return reader_datasets

    def _load_datasets_by_readers(self, reader_datasets, **kwargs):
        # load all datasets for one reader at a time
        loaded_datasets = DatasetDict()
        for reader_name, ds_ids in reader_datasets.items():
            reader_instance = self._readers[reader_name]
            new_datasets = reader_instance.load(ds_ids, **kwargs)
            loaded_datasets.update(new_datasets)
        return loaded_datasets

    def generate_possible_composites(self, unload):
        """See which composites can be generated and generate them.

        Args:
            unload (bool): if the dependencies of the composites
                           should be unloaded after successful generation.
        """
        keepables = self._generate_composites_from_loaded_datasets()

        if self.missing_datasets:
            self._remove_failed_datasets(keepables)
        if unload:
            self.unload(keepables=keepables)

    def _filter_loaded_datasets_from_trunk_nodes(self, trunk_nodes):
        loaded_data_ids = self._datasets.keys()
        for trunk_node in trunk_nodes:
            if trunk_node.name in loaded_data_ids:
                continue
            yield trunk_node

    def _generate_composites_from_loaded_datasets(self):
        """Compute all the composites contained in `requirements`."""
        trunk_nodes = self._dependency_tree.trunk(limit_nodes_to=self.missing_datasets,
                                                  limit_children_to=self._datasets.keys())
        needed_comp_nodes = set(self._filter_loaded_datasets_from_trunk_nodes(trunk_nodes))
        return self._generate_composites_nodes_from_loaded_datasets(needed_comp_nodes)

    def _generate_composites_nodes_from_loaded_datasets(self, compositor_nodes):
        """Read (generate) composites."""
        keepables = set()
        for node in compositor_nodes:
            self._generate_composite(node, keepables)
        return keepables

    def _generate_composite(self, comp_node: CompositorNode, keepables: set):
        """Collect all composite prereqs and create the specified composite.

        Args:
            comp_node: Composite Node to generate a Dataset for
            keepables: `set` to update if any datasets are needed
                       when generation is continued later. This can
                       happen if generation is delayed to incompatible
                       areas which would require resampling first.

        """
        if self._datasets.contains(comp_node.name):
            # already loaded
            return

        compositor = comp_node.compositor
        prereqs = comp_node.required_nodes
        optional_prereqs = comp_node.optional_nodes

        try:
            delayed_prereq = False
            prereq_datasets = self._get_prereq_datasets(
                comp_node.name,
                prereqs,
                keepables,
            )
        except DelayedGeneration:
            # if we are missing a required dependency that could be generated
            # later then we need to wait to return until after we've also
            # processed the optional dependencies
            delayed_prereq = True
        except KeyError:
            # we are missing a hard requirement that will never be available
            # there is no need to "keep" optional dependencies
            return

        optional_datasets = self._get_prereq_datasets(
            comp_node.name,
            optional_prereqs,
            keepables,
            skip=True
        )

        # we are missing some prerequisites
        # in the future we may be able to generate this composite (delayed)
        # so we need to hold on to successfully loaded prerequisites and
        # optional prerequisites
        if delayed_prereq:
            preservable_datasets = set(self._datasets.keys())
            prereq_ids = set(p.name for p in prereqs)
            opt_prereq_ids = set(p.name for p in optional_prereqs)
            keepables |= preservable_datasets & (prereq_ids | opt_prereq_ids)
            return

        try:
            composite = compositor(prereq_datasets,
                                   optional_datasets=optional_datasets,
                                   **comp_node.name.to_dict())
            cid = DataID.new_id_from_dataarray(composite)
            self._datasets[cid] = composite

            # update the node with the computed DataID
            if comp_node.name in self._wishlist:
                self._wishlist.remove(comp_node.name)
                self._wishlist.add(cid)
            self._dependency_tree.update_node_name(comp_node, cid)
        except IncompatibleAreas:
            LOG.debug("Delaying generation of %s because of incompatible areas", str(compositor.id))
            preservable_datasets = set(self._datasets.keys())
            prereq_ids = set(p.name for p in prereqs)
            opt_prereq_ids = set(p.name for p in optional_prereqs)
            keepables |= preservable_datasets & (prereq_ids | opt_prereq_ids)
            # even though it wasn't generated keep a list of what
            # might be needed in other compositors
            keepables.add(comp_node.name)
            return

    def _get_prereq_datasets(self, comp_id, prereq_nodes, keepables, skip=False):
        """Get a composite's prerequisites, generating them if needed.

        Args:
            comp_id (DataID): DataID for the composite whose
                                 prerequisites are being collected.
            prereq_nodes (Sequence[Node]): Prerequisites to collect
            keepables (set): `set` to update if any prerequisites can't
                             be loaded at this time (see
                             `_generate_composite`).
            skip (bool): If True, consider prerequisites as optional and
                         only log when they are missing. If False,
                         prerequisites are considered required and will
                         raise an exception and log a warning if they can't
                         be collected. Defaults to False.

        Raises:
            KeyError: If required (skip=False) prerequisite can't be collected.

        """
        prereq_datasets = []
        delayed_gen = False
        for prereq_node in prereq_nodes:
            prereq_id = prereq_node.name
            if prereq_id not in self._datasets and prereq_id not in keepables \
                    and isinstance(prereq_node, CompositorNode):
                self._generate_composite(prereq_node, keepables)

            # composite generation may have updated the DataID
            prereq_id = prereq_node.name
            if prereq_node is self._dependency_tree.empty_node:
                # empty sentinel node - no need to load it
                continue
            elif prereq_id in self._datasets:
                prereq_datasets.append(self._datasets[prereq_id])
            elif isinstance(prereq_node, CompositorNode) and prereq_id in keepables:
                delayed_gen = True
                continue
            elif not skip:
                LOG.debug("Missing prerequisite for '{}': '{}'".format(
                    comp_id, prereq_id))
                raise KeyError("Missing composite prerequisite for"
                               " '{}': '{}'".format(comp_id, prereq_id))
            else:
                LOG.debug("Missing optional prerequisite for {}: {}".format(comp_id, prereq_id))

        if delayed_gen:
            keepables.add(comp_id)
            keepables.update([x.name for x in prereq_nodes])
            LOG.debug("Delaying generation of %s because of dependency's delayed generation: %s", comp_id, prereq_id)
            if not skip:
                LOG.debug("Delayed prerequisite for '{}': '{}'".format(comp_id, prereq_id))
                raise DelayedGeneration(
                    "Delayed composite prerequisite for "
                    "'{}': '{}'".format(comp_id, prereq_id))
            else:
                LOG.debug("Delayed optional prerequisite for {}: {}".format(comp_id, prereq_id))

        return prereq_datasets



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2015-2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Behave steps related to saving or showing datasets."""

from unittest.mock import patch

from behave import given, then, use_step_matcher, when

use_step_matcher("re")


@given("a dataset is available")
def step_impl_create_scene_one_dataset(context):
    """Create a Scene with a fake dataset for testing.

    Args:
        context (behave.runner.Context): Test context

    """
    from xarray import DataArray

    from satpy import Scene
    scn = Scene()
    scn["MyDataset"] = DataArray([[1, 2], [3, 4]], dims=["y", "x"])
    context.scene = scn


@when("the show command is called")
def step_impl_scene_show(context):
    """Call the Scene.show method.

    Args:
        context (behave.runner.Context): Test context

    """
    with patch("trollimage.xrimage.XRImage.show") as mock_show:
        context.scene.show("MyDataset")
        mock_show.assert_called_once_with()


@then("an image should pop up")
def step_impl_image_pop_up(context):
    """Check that a image window pops up (no-op currently).

    Args:
        context (behave.runner.Context): Test context

    """


@when("the save_dataset command is called")
def step_impl_save_dataset_to_png(context):
    """Run Scene.save_dataset to create a PNG image.

    Args:
        context (behave.runner.Context): Test context

    """
    context.filename = "/tmp/test_dataset.png"
    context.scene.save_dataset("MyDataset", context.filename)


@then("a file should be saved on disk")
def step_impl_file_exists_and_remove(context):
    """Check that a file exists on disk and then remove it.

    Args:
        context (behave.runner.Context): Test context

    """
    import os
    assert os.path.exists(context.filename)
    os.remove(context.filename)


@given("a bunch of datasets are available")
def step_impl_create_scene_two_datasets(context):
    """Create a Scene with two fake datasets for testing.

    Args:
        context (behave.runner.Context): Test context

    """
    from xarray import DataArray

    from satpy import Scene
    scn = Scene()
    scn["MyDataset"] = DataArray([[1, 2], [3, 4]], dims=["y", "x"])
    scn["MyDataset2"] = DataArray([[5, 6], [7, 8]], dims=["y", "x"])
    context.scene = scn


@when("the save_datasets command is called")
def step_impl_save_datasets(context):
    """Run Scene.save_datsets to create PNG images.

    Args:
        context (behave.runner.Context): Test context

    """
    context.scene.save_datasets(writer="simple_image", filename="{name}.png")


@then("a bunch of files should be saved on disk")
def step_impl_check_two_pngs_exist(context):
    """Check that two PNGs exist.

    Args:
        context (behave.runner.Context): Test context

    """
    import os
    for filename in ["MyDataset.png", "MyDataset2.png"]:
        assert os.path.exists(filename)
        os.remove(filename)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2022, 2023 Satpy Developers

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Reader for the ATMS SDR format.

A reader for Advanced Technology Microwave Sounder (ATMS) SDR data as it
e.g. comes out of the CSPP package for processing Direct Readout data.

The format is described in the JPSS COMMON DATA FORMAT CONTROL BOOK (CDFCB):

Joint Polar Satellite System (JPSS) Common Data Format Control Book -
External (CDFCB-X) Volume III - SDR/TDR Formats

(474-00001-03_JPSS-CDFCB-X-Vol-III_0124C.pdf)


https://www.nesdis.noaa.gov/about/documents-reports/jpss-technical-documents/jpss-science-documents

"""

import logging
import os

import dask.array as da
import h5py
import xarray as xr

from satpy.readers.core.viirs_atms_sdr import DATASET_KEYS, JPSS_SDR_FileHandler
from satpy.utils import get_legacy_chunk_size

LOG = logging.getLogger(__name__)

CHUNK_SIZE = get_legacy_chunk_size()
ATMS_CHANNEL_NAMES = ["1", "2", "3", "4", "5", "6", "7", "8", "9", "10",
                      "11", "12", "13", "14", "15", "16", "17", "18", "19", "20", "21", "22"]


class ATMS_SDR_FileHandler(JPSS_SDR_FileHandler):
    """ATMS SDR HDF5 File Reader."""

    def __init__(self, filename, filename_info, filetype_info, **kwargs):
        """Initialize file handler."""
        self.datasets = os.path.basename(filename).split("_")[0].split("-")
        super().__init__(filename, filename_info, filetype_info, **kwargs)

    def __getitem__(self, key):
        """Get item for given key."""
        val = self.file_content[key]
        if isinstance(val, h5py.Dataset):
            dset = h5py.File(self.filename, "r")[key]
            if dset.ndim == 3:
                dset_data = da.from_array(dset, chunks=CHUNK_SIZE)
                attrs = self._attrs_cache.get(key, dset.attrs)
                return xr.DataArray(dset_data, dims=["y", "x", "z"], attrs=attrs)

        return super().__getitem__(key)

    def _get_atms_channel_index(self, ch_name):
        """Get the channels array index from name."""
        try:
            return ATMS_CHANNEL_NAMES.index(ch_name)
        except ValueError:
            return None

    def _get_scans_per_granule(self, dataset_group):
        number_of_granules_path = "Data_Products/{dataset_group}/{dataset_group}_Aggr/attr/AggregateNumberGranules"
        nb_granules_path = number_of_granules_path.format(dataset_group=DATASET_KEYS[dataset_group])
        scans = []
        for granule in range(self[nb_granules_path]):
            scans_path = "Data_Products/{dataset_group}/{dataset_group}_Gran_{granule}/attr/N_Number_Of_Scans"
            scans_path = scans_path.format(dataset_group=DATASET_KEYS[dataset_group], granule=granule)
            scans.append(self[scans_path])
        return scans

    def _get_variable(self, var_path, channel_index=None):
        if channel_index is not None:
            return self[var_path][:, :, channel_index]
        return super()._get_variable(var_path)

    def get_dataset(self, dataset_id, ds_info):
        """Get the dataset corresponding to *dataset_id*.

        The size of the return DataArray will be dependent on the number of
        scans actually sensed of course.

        """
        dataset_group = [ds_group for ds_group in ds_info["dataset_groups"] if ds_group in self.datasets]
        if not dataset_group:
            return

        dataset_group = dataset_group[0]
        ds_info["dataset_group"] = dataset_group
        var_path = self._generate_file_key(dataset_id, ds_info)

        ch_index = self._get_atms_channel_index(ds_info["name"])
        data = self.concatenate_dataset(dataset_group, var_path, channel_index=ch_index)
        data = self.mask_fill_values(data, ds_info)

        data = self.scale_data_to_specified_unit(data, dataset_id, ds_info)
        data = self._update_data_attributes(data, dataset_id, ds_info)

        return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2011-2017 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Classes and utilities for defining generic "plugin" components."""

import logging

import yaml
from yaml import UnsafeLoader

from satpy._config import config_search_paths
from satpy.utils import recursive_dict_update

LOG = logging.getLogger(__name__)


class Plugin(object):
    """Base plugin class for all dynamically loaded and configured objects."""

    def __init__(self, default_config_filename=None, config_files=None, **kwargs):
        """Load configuration files related to this plugin.

        This initializes a `self.config` dictionary that can be used to customize the subclass.

        Args:
            default_config_filename (str): Configuration filename to use if
                no other files have been specified with `config_files`.
            config_files (list or str): Configuration files to load instead
                of those automatically found in `SATPY_CONFIG_PATH` and other
                default configuration locations.
            kwargs (dict): Unused keyword arguments.

        """
        self.default_config_filename = default_config_filename
        self.config_files = config_files
        if self.config_files is None and self.default_config_filename is not None:
            # Specify a default
            self.config_files = config_search_paths(self.default_config_filename)
        if not isinstance(self.config_files, (list, tuple)):
            self.config_files = [self.config_files]

        self.config = {}
        if self.config_files:
            for config_file in self.config_files:
                self.load_yaml_config(config_file)

    def load_yaml_config(self, conf):
        """Load a YAML configuration file and recursively update the overall configuration."""
        with open(conf, "r", encoding="utf-8") as fd:
            recursive_dict_update(self.config, yaml.load(fd, Loader=UnsafeLoader))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
# type: ignore

"""An OSISAF SST reader for the netCDF GHRSST format."""

import datetime as dt
import logging

import numpy as np

from satpy.dataset import Dataset
from satpy.readers.core.netcdf import NetCDF4FileHandler

logger = logging.getLogger(__name__)

PLATFORM_NAME = {"NPP": "Suomi-NPP", }
SENSOR_NAME = {"VIIRS": "viirs",
               "AVHRR": "avhrr/3"}


class GHRSST_OSISAFL2(NetCDF4FileHandler):
    """Reader for the OSISAF SST GHRSST format."""

    def _parse_datetime(self, datestr):
        return dt.datetime.strptime(datestr, "%Y%m%dT%H%M%SZ")

    def get_area_def(self, area_id, area_info):
        """Override abstract baseclass method."""
        raise NotImplementedError

    def get_dataset(self, dataset_id, ds_info, out=None):
        """Load a dataset."""
        var_path = ds_info.get("file_key", "{}".format(dataset_id["name"]))
        dtype = ds_info.get("dtype", np.float32)
        if var_path + "/shape" not in self:
            # loading a scalar value
            shape = 1
        else:
            shape = self[var_path + "/shape"]
            if shape[0] == 1:
                # Remove the time dimenstion from dataset
                shape = shape[1], shape[2]

        file_units = ds_info.get("file_units")
        if file_units is None:
            try:
                file_units = self[var_path + "/attr/units"]
                # they were almost completely CF compliant...
                if file_units == "none":
                    file_units = "1"
            except KeyError:
                # no file units specified
                file_units = None

        if out is None:
            out = np.ma.empty(shape, dtype=dtype)
            out.mask = np.zeros(shape, dtype=bool)

        out.data[:] = np.require(self[var_path][0][::-1], dtype=dtype)
        self._scale_and_mask_data(out, var_path)

        ds_info.update({
            "units": ds_info.get("units", file_units),
            "platform_name": PLATFORM_NAME.get(self["/attr/platform"], self["/attr/platform"]),
            "sensor": SENSOR_NAME.get(self["/attr/sensor"], self["/attr/sensor"]),
        })
        ds_info.update(dataset_id.to_dict())
        cls = ds_info.pop("container", Dataset)
        return cls(out, **ds_info)

    def _scale_and_mask_data(self, out, var_path):
        valid_min = self[var_path + "/attr/valid_min"]
        valid_max = self[var_path + "/attr/valid_max"]
        try:
            scale_factor = self[var_path + "/attr/scale_factor"]
            scale_offset = self[var_path + "/attr/add_offset"]
        except KeyError:
            scale_factor = scale_offset = None
        if valid_min is not None and valid_max is not None:
            out.mask[:] |= (out.data < valid_min) | (out.data > valid_max)
        factors = (scale_factor, scale_offset)
        if factors[0] != 1 or factors[1] != 0:
            out.data[:] *= factors[0]
            out.data[:] += factors[1]

    def get_lonlats(self, navid, nav_info, lon_out=None, lat_out=None):
        """Load an area."""
        lon_key = "lon"
        valid_min = self[lon_key + "/attr/valid_min"]
        valid_max = self[lon_key + "/attr/valid_max"]

        lon_out.data[:] = self[lon_key][::-1]
        lon_out.mask[:] = (lon_out < valid_min) | (lon_out > valid_max)

        lat_key = "lat"
        valid_min = self[lat_key + "/attr/valid_min"]
        valid_max = self[lat_key + "/attr/valid_max"]
        lat_out.data[:] = self[lat_key][::-1]
        lat_out.mask[:] = (lat_out < valid_min) | (lat_out > valid_max)

        return {}

    @property
    def start_time(self):
        """Get start time."""
        # return self.filename_info['start_time']
        return self._parse_datetime(self["/attr/start_time"])

    @property
    def end_time(self):
        """Get end time."""
        return self._parse_datetime(self["/attr/stop_time"])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Satpy developers
#
# This file is part of Satpy.
#
# Satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Satpy.  If not, see <http://www.gnu.org/licenses/>.
#
#

"""Reader for Mimic TPW data in netCDF format from SSEC.

This module implements reader for MIMIC_TPW2 netcdf files.
MIMIC-TPW2 is an experimental global product of total precipitable water (TPW),
using morphological compositing of the MIRS retrieval from several available
operational microwave-frequency sensors. Originally described in a 2010 paper by
Wimmers and Velden. This Version 2 is developed from an older method that uses simpler,
but more limited TPW retrievals and advection calculations.

More information, data and credits at
http://tropic.ssec.wisc.edu/real-time/mtpw2/credits.html
"""

import logging

import numpy as np
import xarray as xr
from pyresample.geometry import AreaDefinition

from satpy.readers.core.netcdf import NetCDF4FileHandler, netCDF4

logger = logging.getLogger(__name__)


class MimicTPW2FileHandler(NetCDF4FileHandler):
    """NetCDF4 reader for MIMC TPW."""

    def __init__(self, filename, filename_info, filetype_info):
        """Initialize the reader."""
        super(MimicTPW2FileHandler, self).__init__(filename, filename_info,
                                                   filetype_info,
                                                   xarray_kwargs={"decode_times": False})

    def available_datasets(self, configured_datasets=None):
        """Get datasets in file matching gelocation shape (lat/lon)."""
        lat_shape = self.file_content.get("/dimension/lat")
        lon_shape = self.file_content.get("/dimension/lon")

        # Read the lat/lon variables?
        handled_variables = set()

        # update previously configured datasets
        logger.debug("Starting previously configured variables loop...")
        for is_avail, ds_info in (configured_datasets or []):
            # some other file handler knows how to load this
            if is_avail is not None:
                yield is_avail, ds_info

            var_name = ds_info.get("file_key", ds_info["name"])
            # logger.debug("Evaluating previously configured variable: %s", var_name)
            matches = self.file_type_matches(ds_info["file_type"])
            # we can confidently say that we can provide this dataset and can
            # provide more info
            if matches and var_name in self:
                logger.debug("Handling previously configured variable: %s", var_name)
                handled_variables.add(var_name)
                new_info = ds_info.copy()  # don't mess up the above yielded
                yield True, new_info
            elif is_avail is None:
                # if we didn't know how to handle this dataset and no one else did
                # then we should keep it going down the chain
                yield is_avail, ds_info

        # Iterate over dataset contents
        for var_name, val in self.file_content.items():
            # Only evaluate variables
            if isinstance(val, netCDF4.Variable):
                logger.debug("Evaluating new variable: %s", var_name)
                var_shape = self[var_name + "/shape"]
                logger.debug("Dims:{}".format(var_shape))
                if var_shape == (lat_shape, lon_shape):
                    logger.debug("Found valid additional dataset: %s", var_name)
                    # Skip anything we have already configured
                    if var_name in handled_variables:
                        logger.debug("Already handled, skipping: %s", var_name)
                        continue
                    handled_variables.add(var_name)
                    # Create new ds_info object
                    new_info = {
                        "name": var_name,
                        "file_key": var_name,
                        "file_type": self.filetype_info["file_type"],
                    }
                    logger.debug(var_name)
                    yield True, new_info

    def get_dataset(self, ds_id, info):
        """Load dataset designated by the given key from file."""
        logger.debug("Getting data for: %s", ds_id["name"])
        file_key = info.get("file_key", ds_id["name"])
        data = np.flipud(self[file_key])
        data = xr.DataArray(data, dims=["y", "x"])
        data.attrs = self.get_metadata(data, info)

        if "lon" in data.dims:
            data.rename({"lon": "x"})
        if "lat" in data.dims:
            data.rename({"lat": "y"})

        return data

    def get_area_def(self, dsid):
        """Flip data up/down and define equirectangular AreaDefintion."""
        flip_lat = np.flipud(self["latArr"])
        latlon = np.meshgrid(self["lonArr"], flip_lat)

        width = self["lonArr/shape"][0]
        height = self["latArr/shape"][0]

        lower_left_x = latlon[0][height-1][0]
        lower_left_y = latlon[1][height-1][0]

        upper_right_y = latlon[1][0][width-1]
        upper_right_x = latlon[0][0][width-1]

        area_extent = (lower_left_x, lower_left_y, upper_right_x, upper_right_y)
        description = "MIMIC TPW WGS84"
        area_id = "mimic"
        proj_id = "World Geodetic System 1984"
        projection = "EPSG:4326"
        area_def = AreaDefinition(area_id, description, proj_id, projection, width, height, area_extent, )
        return area_def

    def get_metadata(self, data, info):
        """Get general metadata for file."""
        metadata = {}
        metadata.update(data.attrs)
        metadata.update(info)
        metadata.update({
            "platform_shortname": "aggregated microwave",
            "sensor": "mimic",
            "start_time": self.start_time,
            "end_time": self.end_time,
        })
        metadata.update(self[info.get("file_key")].variable.attrs)

        return metadata

    @property
    def start_time(self):
        """Start timestamp of the dataset determined from yaml."""
        return self.filename_info["start_time"]

    @property
    def end_time(self):
        """End timestamp of the dataset same as start_time."""
        return self.filename_info.get("end_time", self.start_time)

    @property
    def sensor_name(self):
        """Sensor name."""
        return self["sensor"]



"""Various functions to use with the :meth:`MultiScene.blend <satpy.multiscene._multiscene.MultiScene.blend>` method."""
from __future__ import annotations

from typing import Callable, Iterable, Mapping, Optional, Sequence

import pandas as pd
import xarray as xr
from dask import array as da

from satpy.dataset import combine_metadata


def stack(
        data_arrays: Sequence[xr.DataArray],
        weights: Optional[Sequence[xr.DataArray]] = None,
        blend_type: str = "select_with_weights"
) -> xr.DataArray:
    """Combine a series of datasets in different ways.

    By default, DataArrays are stacked on top of each other, so the last one
    applied is on top. Each DataArray is assumed to represent the same
    geographic region, meaning they have the same area. If a sequence of
    weights is provided then they must have the same shape as the area.
    Weights with greater than 2 dimensions are not currently supported.

    When weights are provided, the DataArrays will be combined according to
    those weights. Data can be integer category products (ex. cloud type),
    single channels (ex. radiance), or a multi-band composite (ex. an RGB or
    RGBA true_color). In the latter case, the weight array is applied
    to each band (R, G, B, A) in the same way. The result will be a composite
    DataArray where each pixel is constructed in a way depending on ``blend_type``.

    Blend type can be one of the following:

     * select_with_weights: The input pixel with the maximum weight is chosen.
     * blend_with_weights: The final pixel is a weighted average of all valid
       input pixels.

    """
    if weights:
        return _stack_with_weights(data_arrays, weights, blend_type)
    return _stack_no_weights(data_arrays)


def _stack_with_weights(
        datasets: Sequence[xr.DataArray],
        weights: Sequence[xr.DataArray],
        blend_type: str
) -> xr.DataArray:
    blend_func = _get_weighted_blending_func(blend_type)
    filled_weights = list(_fill_weights_for_invalid_dataset_pixels(datasets, weights))
    return blend_func(datasets, filled_weights)


def _get_weighted_blending_func(blend_type: str) -> Callable:
    WEIGHTED_BLENDING_FUNCS = {
        "select_with_weights": _stack_select_by_weights,
        "blend_with_weights": _stack_blend_by_weights,
    }
    blend_func = WEIGHTED_BLENDING_FUNCS.get(blend_type)
    if blend_func is None:
        raise ValueError(f"Unknown weighted blending type: {blend_type}."
                         f"Expected one of: {WEIGHTED_BLENDING_FUNCS.keys()}")
    return blend_func


def _fill_weights_for_invalid_dataset_pixels(
        datasets: Sequence[xr.DataArray],
        weights: Sequence[xr.DataArray]
) -> Iterable[xr.DataArray]:
    """Replace weight valus with 0 where data values are invalid/null."""
    has_bands_dims = "bands" in datasets[0].dims
    for i, dataset in enumerate(datasets):
        # if multi-band only use the red-band
        compare_ds = dataset[0] if has_bands_dims else dataset
        try:
            yield xr.where(compare_ds == compare_ds.attrs["_FillValue"], 0, weights[i])
        except KeyError:
            yield xr.where(compare_ds.isnull(), 0, weights[i])


def _stack_blend_by_weights(
        datasets: Sequence[xr.DataArray],
        weights: Sequence[xr.DataArray],
) -> xr.DataArray:
    """Stack datasets blending overlap using weights."""
    attrs = _combine_stacked_attrs([data_arr.attrs for data_arr in datasets])

    overlays = []
    for weight, overlay in zip(weights, datasets):
        # Any 'overlay' fill values should already be reflected in the weights
        # as 0. See _fill_weights_for_invalid_dataset_pixels. We fill NA with
        # 0 here to avoid NaNs affecting valid pixels in other datasets. Note
        # `.fillna` does not handle the `_FillValue` attribute so this filling
        # is purely to remove NaNs.
        overlays.append(overlay.fillna(0) * weight)
    # NOTE: Currently no way to ignore numpy divide by 0 warnings without
    # making a custom map_blocks version of the divide
    base = sum(overlays) / sum(weights)

    dims = datasets[0].dims
    blended_array = xr.DataArray(base, dims=dims, attrs=attrs)
    return blended_array


def _stack_select_by_weights(
        datasets: Sequence[xr.DataArray],
        weights: Sequence[xr.DataArray],
) -> xr.DataArray:
    """Stack datasets selecting pixels using weights."""
    indices = da.argmax(da.dstack(weights), axis=-1)
    if "bands" in datasets[0].dims:
        indices = [indices] * datasets[0].sizes["bands"]

    attrs = _combine_stacked_attrs([data_arr.attrs for data_arr in datasets])
    dims = datasets[0].dims
    coords = datasets[0].coords
    selected_array = xr.DataArray(da.choose(indices, datasets), dims=dims, coords=coords, attrs=attrs)
    return selected_array


def _stack_no_weights(
        datasets: Sequence[xr.DataArray],
) -> xr.DataArray:
    base = datasets[0].copy()
    collected_attrs = [base.attrs]
    for data_arr in datasets[1:]:
        collected_attrs.append(data_arr.attrs)
        try:
            base = base.where(data_arr == data_arr.attrs["_FillValue"], data_arr)
        except KeyError:
            base = base.where(data_arr.isnull(), data_arr)

    attrs = _combine_stacked_attrs(collected_attrs)
    base.attrs = attrs
    return base


def _combine_stacked_attrs(collected_attrs: Sequence[Mapping]) -> dict:
    return combine_metadata(*collected_attrs)


def timeseries(datasets):
    """Expand dataset with and concatenate by time dimension."""
    expanded_ds = []
    for ds in datasets:
        if "time" not in ds.dims:
            tmp = ds.expand_dims("time")
            tmp.coords["time"] = pd.DatetimeIndex([ds.attrs["start_time"]])
        else:
            tmp = ds
        expanded_ds.append(tmp)

    res = xr.concat(expanded_ds, dim="time")
    res.attrs = combine_metadata(*[x.attrs for x in expanded_ds])
    return res


def temporal_rgb(
        data_arrays: Sequence[xr.DataArray],
) -> xr.DataArray:
    """Combine a series of datasets as a temporal RGB.

    The first dataset is used as the Red component of the new composite, the second as Green and the third as Blue.
    All the other datasets are discarded.
    """
    from satpy.composites.core import GenericCompositor

    compositor = GenericCompositor("temporal_composite")
    composite = compositor((data_arrays[0], data_arrays[1], data_arrays[2]))
    composite.attrs = data_arrays[2].attrs

    return composite



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2016-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Module for testing the satpy.readers.olci_nc module."""
import datetime
import unittest
import unittest.mock as mock


class TestOLCIReader(unittest.TestCase):
    """Test various olci_nc filehandlers."""

    @mock.patch("xarray.open_dataset")
    def test_instantiate(self, mocked_dataset):
        """Test initialization of file handlers."""
        import xarray as xr

        from satpy.readers.olci_nc import NCOLCI1B, NCOLCI2, NCOLCIBase, NCOLCICal, NCOLCIChannelBase, NCOLCIGeo
        from satpy.tests.utils import make_dataid

        cal_data = xr.Dataset(
            {
                "solar_flux": (("bands"), [0, 1, 2]),
                "detector_index": (("bands"), [0, 1, 2]),
            },
            {"bands": [0, 1, 2], },
        )

        ds_id = make_dataid(name="Oa01", calibration="reflectance")
        ds_id2 = make_dataid(name="wsqf", calibration="reflectance")
        filename_info = {"mission_id": "S3A", "dataset_name": "Oa01", "start_time": 0, "end_time": 0}

        test = NCOLCIBase("somedir/somefile.nc", filename_info, "c")
        test.get_dataset(ds_id, filename_info)
        mocked_dataset.assert_called()
        mocked_dataset.reset_mock()

        test = NCOLCICal("somedir/somefile.nc", filename_info, "c")
        test.get_dataset(ds_id, filename_info)
        mocked_dataset.assert_called()
        mocked_dataset.reset_mock()

        test = NCOLCIGeo("somedir/somefile.nc", filename_info, "c")
        test.get_dataset(ds_id, filename_info)
        mocked_dataset.assert_called()
        mocked_dataset.reset_mock()

        test = NCOLCIChannelBase("somedir/somefile.nc", filename_info, "c")
        test.get_dataset(ds_id, filename_info)
        mocked_dataset.assert_called()
        mocked_dataset.reset_mock()

        cal = mock.Mock()
        cal.nc = cal_data
        test = NCOLCI1B("somedir/somefile.nc", filename_info, "c", cal)
        test.get_dataset(ds_id, filename_info)
        mocked_dataset.assert_called()
        mocked_dataset.reset_mock()

        test = NCOLCI2("somedir/somefile.nc", filename_info, "c")
        test.get_dataset(ds_id, {"nc_key": "the_key"})
        test.get_dataset(ds_id2, {"nc_key": "the_key"})
        mocked_dataset.assert_called()
        mocked_dataset.reset_mock()

    @mock.patch("xarray.open_dataset")
    def test_open_file_objects(self, mocked_open_dataset):
        """Test initialization of file handlers."""
        from satpy.readers.olci_nc import NCOLCIBase
        filename_info = {"mission_id": "S3A", "dataset_name": "Oa01", "start_time": 0, "end_time": 0}

        open_file = mock.MagicMock()

        file_handler = NCOLCIBase(open_file, filename_info, "c")
        #  deepcode ignore W0104: This is a property that is actually a function call.
        file_handler.nc  # pylint: disable=W0104
        mocked_open_dataset.assert_called()
        open_file.open.assert_called()
        assert (open_file.open.return_value in mocked_open_dataset.call_args[0] or
                open_file.open.return_value == mocked_open_dataset.call_args[1].get("filename_or_obj"))

    @mock.patch("xarray.open_dataset")
    def test_get_l2_mask(self, mocked_dataset):
        """Test reading datasets."""
        import numpy as np
        import xarray as xr

        from satpy.readers.olci_nc import NCOLCI2
        from satpy.tests.utils import make_dataid
        mocked_dataset.return_value = xr.Dataset({"mask": (["rows", "columns"],
                                                           np.array([1 << x for x in range(30)]).reshape(5, 6))},
                                                 coords={"rows": np.arange(5),
                                                         "columns": np.arange(6)})
        ds_id = make_dataid(name="mask")
        filename_info = {"mission_id": "S3A", "dataset_name": "mask", "start_time": 0, "end_time": 0}
        test = NCOLCI2("somedir/somefile.nc", filename_info, "c")
        res = test.get_dataset(ds_id, {"nc_key": "mask"})
        assert res.dtype == np.dtype("bool")
        expected = np.array([[True, False, True, True, True, True],
                             [False, False, True, True, False, False],
                             [False, False, False, False, False, True],
                             [False, True, False, False, False, True],
                             [True, False, False, True, False, False]])
        np.testing.assert_array_equal(res.values, expected)

    @mock.patch("xarray.open_dataset")
    def test_get_l2_mask_with_alternative_items(self, mocked_dataset):
        """Test reading datasets."""
        import numpy as np
        import xarray as xr

        from satpy.readers.olci_nc import NCOLCI2
        from satpy.tests.utils import make_dataid
        mocked_dataset.return_value = xr.Dataset({"mask": (["rows", "columns"],
                                                           np.array([1 << x for x in range(30)]).reshape(5, 6))},
                                                 coords={"rows": np.arange(5),
                                                         "columns": np.arange(6)})
        ds_id = make_dataid(name="mask")
        filename_info = {"mission_id": "S3A", "dataset_name": "mask", "start_time": 0, "end_time": 0}
        test = NCOLCI2("somedir/somefile.nc", filename_info, "c", mask_items=["INVALID"])
        res = test.get_dataset(ds_id, {"nc_key": "mask"})
        assert res.dtype == np.dtype("bool")
        expected = np.array([True] + [False] * 29).reshape(5, 6)
        np.testing.assert_array_equal(res.values, expected)


    @mock.patch("xarray.open_dataset")
    def test_get_l1b_default_mask(self, mocked_dataset):
        """Test reading mask datasets from L1B products."""
        import numpy as np
        import xarray as xr

        from satpy.readers.olci_nc import NCOLCI1B
        from satpy.tests.utils import make_dataid
        mocked_dataset.return_value = xr.Dataset({"quality_flags": (["rows", "columns"],
                                                           np.array([1 << (x % 32) for x in range(35)]).reshape(5, 7))},
                                                 coords={"rows": np.arange(5),
                                                         "columns": np.arange(7)})
        ds_id = make_dataid(name="mask")
        filename_info = {"mission_id": "S3A", "dataset_name": "mask", "start_time": 0, "end_time": 0}
        test = NCOLCI1B("somedir/somefile.nc", filename_info, "c")
        res = test.get_dataset(ds_id, {"nc_key": "quality_flags"})
        assert res.dtype == np.dtype("bool")

        expected = np.array([[False, False, False, False, False, False, False],
                             [False, False, False, False, False, False, False],
                             [False, False, False, False, False, False, False],
                             [True, True, True, True, True, True, True],
                             [True, False, True, True, False, False, False]])

        np.testing.assert_array_equal(res.values, expected)


    @mock.patch("xarray.open_dataset")
    def test_get_l1b_customized_mask(self, mocked_dataset):
        """Test reading mask datasets from L1B products."""
        import numpy as np
        import xarray as xr

        from satpy.readers.olci_nc import NCOLCI1B
        from satpy.tests.utils import make_dataid
        mocked_dataset.return_value = xr.Dataset({"quality_flags": (["rows", "columns"],
                                                           np.array([1 << (x % 32) for x in range(35)]).reshape(5, 7))},
                                                 coords={"rows": np.arange(5),
                                                         "columns": np.arange(7)})
        ds_id = make_dataid(name="mask")
        filename_info = {"mission_id": "S3A", "dataset_name": "mask", "start_time": 0, "end_time": 0}
        test = NCOLCI1B("somedir/somefile.nc", filename_info, "c", mask_items=["bright", "invalid"])
        res = test.get_dataset(ds_id, {"nc_key": "quality_flags"})
        assert res.dtype == np.dtype("bool")

        expected = np.array([[False, False, False, False, False, False, False],
                             [False, False, False, False, False, False, False],
                             [False, False, False, False, False, False, False],
                             [False, False, False, False, True, False, True],
                             [False, False, False, False, False, False, False]])

        np.testing.assert_array_equal(res.values, expected)


    @mock.patch("xarray.open_dataset")
    def test_olci_angles(self, mocked_dataset):
        """Test reading datasets."""
        import numpy as np
        import xarray as xr

        from satpy.readers.olci_nc import NCOLCIAngles
        from satpy.tests.utils import make_dataid
        attr_dict = {
            "ac_subsampling_factor": 1,
            "al_subsampling_factor": 2,
        }
        mocked_dataset.return_value = xr.Dataset({"SAA": (["tie_rows", "tie_columns"],
                                                          np.array([1 << x for x in range(30)]).reshape(5, 6)),
                                                  "SZA": (["tie_rows", "tie_columns"],
                                                          np.array([1 << x for x in range(30)]).reshape(5, 6)),
                                                  "OAA": (["tie_rows", "tie_columns"],
                                                          np.array([1 << x for x in range(30)]).reshape(5, 6)),
                                                  "OZA": (["tie_rows", "tie_columns"],
                                                          np.array([1 << x for x in range(30)]).reshape(5, 6))},
                                                 coords={"rows": np.arange(5),
                                                         "columns": np.arange(6)},
                                                 attrs=attr_dict)
        filename_info = {"mission_id": "S3A", "dataset_name": "Oa01", "start_time": 0, "end_time": 0}

        ds_id = make_dataid(name="solar_azimuth_angle")
        ds_id2 = make_dataid(name="satellite_zenith_angle")
        test = NCOLCIAngles("somedir/somefile.nc", filename_info, "c")
        test.get_dataset(ds_id, filename_info)
        test.get_dataset(ds_id2, filename_info)
        mocked_dataset.assert_called()
        mocked_dataset.reset_mock()

    @mock.patch("xarray.open_dataset")
    def test_olci_meteo(self, mocked_dataset):
        """Test reading datasets."""
        import numpy as np
        import xarray as xr

        from satpy.readers.olci_nc import NCOLCIMeteo
        from satpy.tests.utils import make_dataid
        attr_dict = {
            "ac_subsampling_factor": 1,
            "al_subsampling_factor": 2,
        }
        data = {"humidity": (["tie_rows", "tie_columns"],
                             np.array([1 << x for x in range(30)]).reshape(5, 6)),
                "total_ozone": (["tie_rows", "tie_columns"],
                                np.array([1 << x for x in range(30)]).reshape(5, 6)),
                "sea_level_pressure": (["tie_rows", "tie_columns"],
                                       np.array([1 << x for x in range(30)]).reshape(5, 6)),
                "total_columnar_water_vapour": (["tie_rows", "tie_columns"],
                                                np.array([1 << x for x in range(30)]).reshape(5, 6))}
        mocked_dataset.return_value = xr.Dataset(data,
                                                 coords={"rows": np.arange(5),
                                                         "columns": np.arange(6)},
                                                 attrs=attr_dict)
        filename_info = {"mission_id": "S3A", "dataset_name": "humidity", "start_time": 0, "end_time": 0}

        ds_id = make_dataid(name="humidity")
        ds_id2 = make_dataid(name="total_ozone")
        test = NCOLCIMeteo("somedir/somefile.nc", filename_info, "c")
        test.get_dataset(ds_id, filename_info)
        test.get_dataset(ds_id2, filename_info)
        mocked_dataset.assert_called()
        mocked_dataset.reset_mock()

    @mock.patch("xarray.open_dataset")
    def test_chl_nn(self, mocked_dataset):
        """Test unlogging the chl_nn product."""
        import numpy as np
        import xarray as xr

        from satpy.readers.olci_nc import NCOLCI2
        from satpy.tests.utils import make_dataid
        attr_dict = {
            "ac_subsampling_factor": 64,
            "al_subsampling_factor": 1,
        }
        data = {"CHL_NN": (["rows", "columns"],
                           np.arange(30).reshape(5, 6).astype(float),
                           {"units": "lg(re mg.m-3)"})}
        mocked_dataset.return_value = xr.Dataset(data,
                                                 coords={"rows": np.arange(5),
                                                         "columns": np.arange(6)},
                                                 attrs=attr_dict)
        ds_info = {"name": "chl_nn", "sensor": "olci", "resolution": 300,
                   "standard_name": "algal_pigment_concentration", "units": "lg(re mg.m-3)",
                   "coordinates": ("longitude", "latitude"), "file_type": "esa_l2_chl_nn", "nc_key": "CHL_NN",
                   "modifiers": ()}
        filename_info = {"mission_id": "S3A", "datatype_id": "WFR",
                         "start_time": datetime.datetime(2019, 9, 24, 9, 29, 39),
                         "end_time": datetime.datetime(2019, 9, 24, 9, 32, 39),
                         "creation_time": datetime.datetime(2019, 9, 24, 11, 40, 26), "duration": 179, "cycle": 49,
                         "relative_orbit": 307, "frame": 1800, "centre": "MAR", "mode": "O", "timeliness": "NR",
                         "collection": "002"}
        ds_id = make_dataid(name="chl_nn")
        file_handler = NCOLCI2("somedir/somefile.nc", filename_info, None, unlog=True)
        res = file_handler.get_dataset(ds_id, ds_info)

        assert res.attrs["units"] == "mg.m-3"
        assert res.values[-1, -1] == 1e29


class TestL2BitFlags(unittest.TestCase):
    """Test the bitflag reading."""

    def test_bitflags(self):
        """Test the BitFlags class."""
        from functools import reduce

        import numpy as np

        from satpy.readers.olci_nc import BitFlags
        flag_list = ["INVALID", "WATER", "LAND", "CLOUD", "SNOW_ICE",
                     "INLAND_WATER", "TIDAL", "COSMETIC", "SUSPECT", "HISOLZEN",
                     "SATURATED", "MEGLINT", "HIGHGLINT", "WHITECAPS",
                     "ADJAC", "WV_FAIL", "PAR_FAIL", "AC_FAIL", "OC4ME_FAIL",
                     "OCNN_FAIL", "Extra_1", "KDM_FAIL", "Extra_2",
                     "CLOUD_AMBIGUOUS", "CLOUD_MARGIN", "BPAC_ON",
                     "WHITE_SCATT", "LOWRW", "HIGHRW"]

        bits = np.array([1 << x for x in range(len(flag_list))])

        bflags = BitFlags(bits)

        items = ["INVALID", "SNOW_ICE", "INLAND_WATER", "SUSPECT",
                 "AC_FAIL", "CLOUD", "HISOLZEN", "OCNN_FAIL",
                 "CLOUD_MARGIN", "CLOUD_AMBIGUOUS", "LOWRW", "LAND"]

        mask = reduce(np.logical_or, [bflags[item] for item in items])
        expected = np.array([True, False, True, True, True, True, False,
                             False, True, True, False, False, False, False,
                             False, False, False, True, False, True, False,
                             False, False, True, True, False, False, True,
                             False])
        assert all(mask == expected)

    def test_bitflags_with_flags_from_array(self):
        """Test reading bitflags from DataArray attributes."""
        from functools import reduce

        import numpy as np
        import xarray as xr

        from satpy.readers.olci_nc import BitFlags

        flag_masks = [1, 2, 4, 8, 4194304, 8388608, 16777216, 16, 32, 64, 128, 256, 512, 1024, 2048, 4096, 8192, 16384,
                      32768, 65536, 131072, 262144, 524288, 1048576, 2097152, 33554432, 67108864, 134217728, 268435456,
                      536870912, 4294967296, 8589934592, 17179869184, 34359738368, 68719476736, 137438953472,
                      274877906944, 549755813888, 1099511627776, 2199023255552, 4398046511104, 8796093022208,
                      17592186044416, 35184372088832, 70368744177664, 140737488355328, 281474976710656, 562949953421312,
                      1125899906842624, 2251799813685248, 4503599627370496, 9007199254740992, 18014398509481984,
                      36028797018963968]
        flag_meanings = ("INVALID WATER LAND CLOUD TURBID_ATM CLOUD_AMBIGUOUS CLOUD_MARGIN SNOW_ICE INLAND_WATER "
                         "COASTLINE TIDAL COSMETIC SUSPECT HISOLZEN SATURATED MEGLINT HIGHGLINT WHITECAPS ADJAC "
                         "WV_FAIL PAR_FAIL AC_FAIL OC4ME_FAIL OCNN_FAIL KDM_FAIL BPAC_ON WHITE_SCATT LOWRW HIGHRW "
                         "IOP_LSD_FAIL ANNOT_ANGSTROM ANNOT_AERO_B ANNOT_ABSO_D ANNOT_ACLIM ANNOT_ABSOA ANNOT_MIXR1 "
                         "ANNOT_DROUT ANNOT_TAU06 RWNEG_O1 RWNEG_O2 RWNEG_O3 RWNEG_O4 RWNEG_O5 RWNEG_O6 RWNEG_O7 "
                         "RWNEG_O8 RWNEG_O9 RWNEG_O10 RWNEG_O11 RWNEG_O12 RWNEG_O16 RWNEG_O17 RWNEG_O18 RWNEG_O21")

        bits = np.array([1 << x for x in range(int(np.log2(max(flag_masks))) + 1)])
        bits_array = xr.DataArray(bits, attrs=dict(flag_masks=flag_masks, flag_meanings=flag_meanings))
        bflags = BitFlags(bits_array)

        items = ["INVALID", "TURBID_ATM"]
        mask = reduce(np.logical_or, [bflags[item] for item in items])

        assert mask[0].item() is True
        assert any(mask[1:22]) is False
        assert mask[22].item() is True
        assert any(mask[23:]) is False

    def test_bitflags_with_dataarray_without_flags(self):
        """Test the BitFlags class."""
        from functools import reduce

        import numpy as np
        import xarray as xr

        from satpy.readers.olci_nc import BitFlags
        flag_list = ["INVALID", "WATER", "LAND", "CLOUD", "SNOW_ICE",
                     "INLAND_WATER", "TIDAL", "COSMETIC", "SUSPECT", "HISOLZEN",
                     "SATURATED", "MEGLINT", "HIGHGLINT", "WHITECAPS",
                     "ADJAC", "WV_FAIL", "PAR_FAIL", "AC_FAIL", "OC4ME_FAIL",
                     "OCNN_FAIL", "Extra_1", "KDM_FAIL", "Extra_2",
                     "CLOUD_AMBIGUOUS", "CLOUD_MARGIN", "BPAC_ON",
                     "WHITE_SCATT", "LOWRW", "HIGHRW"]

        bits = np.array([1 << x for x in range(len(flag_list))])

        bflags = BitFlags(xr.DataArray(bits))

        items = ["INVALID", "SNOW_ICE", "INLAND_WATER", "SUSPECT",
                 "AC_FAIL", "CLOUD", "HISOLZEN", "OCNN_FAIL",
                 "CLOUD_MARGIN", "CLOUD_AMBIGUOUS", "LOWRW", "LAND"]

        mask = reduce(np.logical_or, [bflags[item] for item in items])
        expected = np.array([True, False, True, True, True, True, False,
                             False, True, True, False, False, False, False,
                             False, False, False, True, False, True, False,
                             False, False, True, True, False, False, True,
                             False])
        assert all(mask == expected)


    def test_bitflags_with_custom_flag_list(self):
        """Test the BitFlags class providing a flag list."""
        from functools import reduce

        import numpy as np

        from satpy.readers.olci_nc import BitFlags
        flag_list = ["INVALID", "WATER", "LAND", "CLOUD", "SNOW_ICE",
                     "INLAND_WATER", "TIDAL", "COSMETIC", "SUSPECT", "HISOLZEN",
                     "SATURATED", "MEGLINT", "HIGHGLINT", "WHITECAPS",
                     "ADJAC", "WV_FAIL", "PAR_FAIL", "AC_FAIL", "OC4ME_FAIL",
                     "OCNN_FAIL", "Extra_1", "KDM_FAIL", "Extra_2",
                     "CLOUD_AMBIGUOUS", "CLOUD_MARGIN", "BPAC_ON",
                     "WHITE_SCATT", "LOWRW", "HIGHRW"]

        bits = np.array([1 << x for x in range(len(flag_list))])

        bflags = BitFlags(bits, flag_list)

        items = ["INVALID", "SNOW_ICE", "INLAND_WATER", "SUSPECT",
                 "AC_FAIL", "CLOUD", "HISOLZEN", "OCNN_FAIL",
                 "CLOUD_MARGIN", "CLOUD_AMBIGUOUS", "LOWRW", "LAND"]

        mask = reduce(np.logical_or, [bflags[item] for item in items])
        expected = np.array([True, False, True, True, True, True, False,
                             False, True, True, False, False, False, False,
                             False, False, False, True, False, True, False,
                             False, False, True, True, False, False, True,
                             False])
        assert all(mask == expected)


class TestL1bBitFlags(unittest.TestCase):
    """Test the bitflag reading."""

    def test_bitflags(self):
        """Test the BitFlags class."""
        from functools import reduce

        import numpy as np

        from satpy.readers.olci_nc import BitFlags


        L1B_QUALITY_FLAGS = ["saturated@Oa21", "saturated@Oa20", "saturated@Oa19", "saturated@Oa18",
                             "saturated@Oa17", "saturated@Oa16", "saturated@Oa15", "saturated@Oa14",
                             "saturated@Oa13", "saturated@Oa12", "saturated@Oa11", "saturated@Oa10",
                             "saturated@Oa09", "saturated@Oa08", "saturated@Oa07", "saturated@Oa06",
                             "saturated@Oa05", "saturated@Oa04", "saturated@Oa03", "saturated@Oa02",
                             "saturated@Oa01", "dubious", "sun-glint_risk", "duplicated",
                             "cosmetic", "invalid", "straylight_risk", "bright",
                             "tidal_region", "fresh_inland_water", "coastline", "land"]

        DEFAULT_L1B_MASK_ITEMS = ["dubious", "sun-glint_risk", "duplicated", "cosmetic", "invalid",
                                  "straylight_risk", "bright", "tidal_region", "coastline", "land"]

        bits = np.array([1 << x for x in range(len(L1B_QUALITY_FLAGS))])

        bflags = BitFlags(bits, flag_list=L1B_QUALITY_FLAGS)

        mask = reduce(np.logical_or, [bflags[item] for item in DEFAULT_L1B_MASK_ITEMS])

        expected = np.array([False, False, False, False,
                             False, False, False, False,
                             False, False, False, False,
                             False, False, False, False,
                             False, False, False, False,
                             False, True, True, True,
                             True, True, True, True,
                             True, False, True, True,
                            ])
        assert all(mask == expected)



"""Resamplers based on the kdtree algorightm."""

import os
import warnings
from logging import getLogger

import dask.array as da
import numpy as np
import xarray as xr
import zarr
from pyresample.resampler import BaseResampler as PRBaseResampler

from satpy.resample.base import _update_resampled_coords
from satpy.utils import get_legacy_chunk_size

LOG = getLogger(__name__)

CHUNK_SIZE = get_legacy_chunk_size()

NN_COORDINATES = {"valid_input_index": ("y1", "x1"),
                  "valid_output_index": ("y2", "x2"),
                  "index_array": ("y2", "x2", "z2")}
BIL_COORDINATES = {"bilinear_s": ("x1", ),
                   "bilinear_t": ("x1", ),
                   "slices_x": ("x1", "n"),
                   "slices_y": ("x1", "n"),
                   "mask_slices": ("x1", "n"),
                   "out_coords_x": ("x2", ),
                   "out_coords_y": ("y2", )}

class KDTreeResampler(PRBaseResampler):
    """Resample using a KDTree-based nearest neighbor algorithm.

    This resampler implements on-disk caching when the `cache_dir` argument
    is provided to the `resample` method. This should provide significant
    performance improvements on consecutive resampling of geostationary data.
    It is not recommended to provide `cache_dir` when the `mask` keyword
    argument is provided to `precompute` which occurs by default for
    `SwathDefinition` source areas.

    Args:
        cache_dir (str): Long term storage directory for intermediate
                         results.
        mask (bool): Force resampled data's invalid pixel mask to be used
                     when searching for nearest neighbor pixels. By
                     default this is True for SwathDefinition source
                     areas and False for all other area definition types.
        radius_of_influence (float): Search radius cut off distance in meters
        epsilon (float): Allowed uncertainty in meters. Increasing uncertainty
                         reduces execution time.

    """

    def __init__(self, source_geo_def, target_geo_def):
        """Init KDTreeResampler."""
        super(KDTreeResampler, self).__init__(source_geo_def, target_geo_def)
        self.resampler = None
        self._index_caches = {}

    def precompute(self, mask=None, radius_of_influence=None, epsilon=0,
                   cache_dir=None, **kwargs):
        """Create a KDTree structure and store it for later use.

        Note: The `mask` keyword should be provided if geolocation may be valid
        where data points are invalid.

        """
        from pyresample.kd_tree import XArrayResamplerNN
        del kwargs
        if mask is not None and cache_dir is not None:
            LOG.warning("Mask and cache_dir both provided to nearest "
                        "resampler. Cached parameters are affected by "
                        "masked pixels. Will not cache results.")
            cache_dir = None

        if radius_of_influence is None and not hasattr(self.source_geo_def, "geocentric_resolution"):
            radius_of_influence = self._adjust_radius_of_influence(radius_of_influence)

        kwargs = dict(source_geo_def=self.source_geo_def,
                      target_geo_def=self.target_geo_def,
                      radius_of_influence=radius_of_influence,
                      neighbours=1,
                      epsilon=epsilon)

        if self.resampler is None:
            # FIXME: We need to move all of this caching logic to pyresample
            self.resampler = XArrayResamplerNN(**kwargs)

        try:
            self.load_neighbour_info(cache_dir, mask=mask, **kwargs)
            LOG.debug("Read pre-computed kd-tree parameters")
        except IOError:
            LOG.debug("Computing kd-tree parameters")
            self.resampler.get_neighbour_info(mask=mask)
            self.save_neighbour_info(cache_dir, mask=mask, **kwargs)

    def _adjust_radius_of_influence(self, radius_of_influence):
        """Adjust radius of influence."""
        warnings.warn(
            "Upgrade 'pyresample' for a more accurate default 'radius_of_influence'.",
            stacklevel=3
        )
        try:
            radius_of_influence = self.source_geo_def.lons.resolution * 3
        except AttributeError:
            try:
                radius_of_influence = max(abs(self.source_geo_def.pixel_size_x),
                                          abs(self.source_geo_def.pixel_size_y)) * 3
            except AttributeError:
                radius_of_influence = 1000

        except TypeError:
            radius_of_influence = 10000
        return radius_of_influence

    def _apply_cached_index(self, val, idx_name, persist=False):
        """Reassign resampler index attributes."""
        if isinstance(val, np.ndarray):
            val = da.from_array(val, chunks=CHUNK_SIZE)
        elif persist and isinstance(val, da.Array):
            val = val.persist()
        setattr(self.resampler, idx_name, val)
        return val

    def load_neighbour_info(self, cache_dir, mask=None, **kwargs):
        """Read index arrays from either the in-memory or disk cache."""
        mask_name = getattr(mask, "name", None)
        cached = {}
        for idx_name in NN_COORDINATES:
            if mask_name in self._index_caches:
                cached[idx_name] = self._apply_cached_index(
                    self._index_caches[mask_name][idx_name], idx_name)
            elif cache_dir:
                cache = self._load_neighbour_info_from_cache(
                    cache_dir, idx_name, mask_name, **kwargs)
                cache = self._apply_cached_index(cache, idx_name)
                cached[idx_name] = cache
            else:
                raise IOError
        self._index_caches[mask_name] = cached

    def _load_neighbour_info_from_cache(self, cache_dir, idx_name, mask_name, **kwargs):
        try:
            filename = self._create_cache_filename(
                cache_dir, prefix="nn_lut-",
                mask=mask_name, **kwargs)
            fid = zarr.open(filename, mode="r")
            cache = np.array(fid[idx_name])
            if idx_name == "valid_input_index":
                # valid input index array needs to be boolean
                cache = cache.astype(bool)
        except ValueError:
            raise IOError
        return cache

    def save_neighbour_info(self, cache_dir, mask=None, **kwargs):
        """Cache resampler's index arrays if there is a cache dir."""
        if cache_dir:
            mask_name = getattr(mask, "name", None)
            cache = self._read_resampler_attrs()
            filename = self._create_cache_filename(
                cache_dir, prefix="nn_lut-", mask=mask_name, **kwargs)
            LOG.info("Saving kd_tree neighbour info to %s", filename)
            zarr_out = xr.Dataset()
            for idx_name, coord in NN_COORDINATES.items():
                # update the cache in place with persisted dask arrays
                cache[idx_name] = self._apply_cached_index(cache[idx_name],
                                                           idx_name,
                                                           persist=True)
                zarr_out[idx_name] = (coord, cache[idx_name])

            # Write indices to Zarr file
            zarr_out.to_zarr(filename)

            self._index_caches[mask_name] = cache
            # Delete the kdtree, it's not needed anymore
            self.resampler.delayed_kdtree = None

    def _read_resampler_attrs(self):
        """Read certain attributes from the resampler for caching."""
        return {attr_name: getattr(self.resampler, attr_name)
                for attr_name in NN_COORDINATES}

    def compute(self, data, weight_funcs=None, fill_value=np.nan,
                with_uncert=False, **kwargs):
        """Resample data."""
        del kwargs
        LOG.debug("Resampling %s", str(data.name))
        res = self.resampler.get_sample_from_neighbour_info(data, fill_value)
        return _update_resampled_coords(data, res, self.target_geo_def)


class BilinearResampler(PRBaseResampler):
    """Resample using bilinear interpolation.

    This resampler implements on-disk caching when the `cache_dir` argument
    is provided to the `resample` method. This should provide significant
    performance improvements on consecutive resampling of geostationary data.

    Args:
        cache_dir (str): Long term storage directory for intermediate
                         results.
        radius_of_influence (float): Search radius cut off distance in meters
        epsilon (float): Allowed uncertainty in meters. Increasing uncertainty
                         reduces execution time.
        reduce_data (bool): Reduce the input data to (roughly) match the
                            target area.

    """

    def __init__(self, source_geo_def, target_geo_def):
        """Init BilinearResampler."""
        super(BilinearResampler, self).__init__(source_geo_def, target_geo_def)
        self.resampler = None

    def precompute(self, mask=None, radius_of_influence=50000, epsilon=0,
                   reduce_data=True, cache_dir=False, **kwargs):
        """Create bilinear coefficients and store them for later use."""
        try:
            from pyresample.bilinear import XArrayBilinearResampler
        except ImportError:
            from pyresample.bilinear import XArrayResamplerBilinear as XArrayBilinearResampler

        del kwargs
        del mask

        if self.resampler is None:
            kwargs = dict(source_geo_def=self.source_geo_def,
                          target_geo_def=self.target_geo_def,
                          radius_of_influence=radius_of_influence,
                          neighbours=32,
                          epsilon=epsilon)

            self.resampler = XArrayBilinearResampler(**kwargs)
            try:
                self.load_bil_info(cache_dir, **kwargs)
                LOG.debug("Loaded bilinear parameters")
            except IOError:
                LOG.debug("Computing bilinear parameters")
                self.resampler.get_bil_info()
                LOG.debug("Saving bilinear parameters.")
                self.save_bil_info(cache_dir, **kwargs)

    def load_bil_info(self, cache_dir, **kwargs):
        """Load bilinear resampling info from cache directory."""
        if cache_dir:
            filename = self._create_cache_filename(cache_dir,
                                                   prefix="bil_lut-",
                                                   **kwargs)
            try:
                self.resampler.load_resampling_info(filename)
            except AttributeError:
                warnings.warn(
                    "Bilinear resampler can't handle caching, "
                    "please upgrade Pyresample to 0.17.0 or newer.",
                    stacklevel=2
                )
                raise IOError
        else:
            raise IOError

    def save_bil_info(self, cache_dir, **kwargs):
        """Save bilinear resampling info to cache directory."""
        if cache_dir:
            filename = self._create_cache_filename(cache_dir,
                                                   prefix="bil_lut-",
                                                   **kwargs)
            # There are some old caches, move them out of the way
            if os.path.exists(filename):
                _move_existing_caches(cache_dir, filename)
            LOG.info("Saving BIL neighbour info to %s", filename)
            try:
                self.resampler.save_resampling_info(filename)
            except AttributeError:
                warnings.warn(
                    "Bilinear resampler can't handle caching, "
                    "please upgrade Pyresample to 0.17.0 or newer.",
                    stacklevel=2
                )

    def compute(self, data, fill_value=None, **kwargs):
        """Resample the given data using bilinear interpolation."""
        del kwargs

        if fill_value is None:
            fill_value = data.attrs.get("_FillValue")
        target_shape = self.target_geo_def.shape

        res = self.resampler.get_sample_from_bil_info(data,
                                                      fill_value=fill_value,
                                                      output_shape=target_shape)

        return _update_resampled_coords(data, res, self.target_geo_def)


def _move_existing_caches(cache_dir, filename):
    """Move existing cache files out of the way."""
    import os
    import shutil
    old_cache_dir = os.path.join(cache_dir, "moved_by_satpy")
    try:
        os.makedirs(old_cache_dir)
    except FileExistsError:
        pass
    try:
        shutil.move(filename, old_cache_dir)
    except shutil.Error:
        os.remove(os.path.join(old_cache_dir,
                               os.path.basename(filename)))
        shutil.move(filename, old_cache_dir)
    LOG.warning("Old cache file was moved to %s", old_cache_dir)


def get_resampler_classes():
    """Get resampler classes based on kdtree."""
    return {
        "kd_tree": KDTreeResampler,
        "nearest": KDTreeResampler,
        "bilinear": BilinearResampler
    }
